# -*- coding: utf-8 -*-

"""
@date: 2020/12/30 下午4:28
@file: mnasnet_unit.py
@author: zj
@description:
"""
from abc import ABC

import copy
import torch.nn as nn

from zcls.model import registry
from zcls.model.conv_helper import get_conv
from zcls.model.norm_helper import get_norm
from zcls.model.act_helper import get_act
from ..misc import round_to_multiple_of
from .mobilenetv3_unit import MobileNetV3Unit

arch_settings = {
    'mobilenetv3-large': [16, 960, 1280,
                          [
                              # kernel_size, stride, inner_planes, with_attention, non-linearity, out_planes
                              [3, 1, 16, 0, 'RE', 16],
                              [3, 2, 64, 0, 'RE', 24],
                              [3, 1, 72, 0, 'RE', 24],
                              [5, 2, 72, 1, 'RE', 40],
                              [5, 1, 120, 1, 'RE', 40],
                              [5, 1, 120, 1, 'RE', 40],
                              [3, 2, 240, 0, 'HS', 80],
                              [3, 1, 200, 0, 'HS', 80],
                              [3, 1, 184, 0, 'HS', 80],
                              [3, 1, 184, 0, 'HS', 80],
                              [3, 1, 480, 1, 'HS', 112],
                              [3, 1, 672, 1, 'HS', 112],
                              [5, 2, 672, 1, 'HS', 160],
                              [5, 1, 960, 1, 'HS', 160],
                              [5, 1, 960, 1, 'HS', 160],
                          ]],
    'mobilenetv3-small': [16, 576, 1024,
                          [
                              # kernel_size, stride, inner_planes, with_attention, non-linearity, out_planes
                              [3, 2, 16, 1, 'RE', 16],
                              [3, 2, 72, 0, 'RE', 24],
                              [3, 1, 88, 0, 'RE', 24],
                              [5, 2, 96, 1, 'HS', 40],
                              [5, 1, 240, 1, 'HS', 40],
                              [5, 1, 240, 1, 'HS', 40],
                              [5, 1, 120, 1, 'HS', 48],
                              [5, 1, 144, 1, 'HS', 48],
                              [5, 2, 288, 1, 'HS', 96],
                              [5, 1, 576, 1, 'HS', 96],
                              [5, 1, 576, 1, 'HS', 96]
                          ]]
}


def relu_or_hswish(name):
    if name == 'RE':
        return nn.ReLU
    elif name == 'HS':
        return nn.Hardswish
    else:
        raise IOError(f'{name} does not exist')


def make_stem(in_planes,
              base_planes,
              inner_planes,
              out_planes,
              conv_layer,
              norm_layer,
              act_layer):
    first_stem = nn.Sequential(
        conv_layer(in_planes, base_planes, kernel_size=3, stride=2, padding=1, bias=False),
        norm_layer(base_planes),
        act_layer(inplace=True)
    )

    last_stem = nn.Sequential(
        conv_layer(inner_planes, out_planes, kernel_size=1, stride=1, padding=0, bias=False),
        norm_layer(out_planes),
        act_layer(inplace=True)
    )

    return first_stem, last_stem


class MobileNetV3Backbone(nn.Module, ABC):

    def __init__(self,
                 in_channels=3,
                 base_channels=16,
                 out_channels=960,
                 layer_setting=None,
                 width_multiplier=1.,
                 round_nearest=8,
                 with_attention=True,
                 reduction=4,
                 attention_type='SqueezeAndExcitationBlock2D',
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None,
                 sigmoid_type=None
                 ):
        """
        :param in_channels: 输入通道数
        :param base_channels: 基础通道数
        :param out_channels: 输出通道数
        :param width_multiplier: 宽度乘法器
        :param round_nearest: 设置每一层通道数均为8的倍数
        :param with_attention: 是否使用注意力模块
        :param reduction: 衰减率
        :param attention_type: 注意力模块类型
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        :param sigmoid_type: sigmoid类型
        """
        super(MobileNetV3Backbone, self).__init__()

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.Hardswish
        if sigmoid_type is None:
            sigmoid_type = 'HSigmoid'
        block_layer = MobileNetV3Unit

        if layer_setting is None:
            layer_setting = [
                # kernel_size, stride, inner_planes, with_attention_2, non-linearity, out_planes
                [3, 1, 16, 0, 'RE', 16],
                [3, 2, 64, 0, 'RE', 24],
                [3, 1, 72, 0, 'RE', 24],
                [5, 2, 72, 1, 'RE', 40],
                [5, 1, 120, 1, 'RE', 40],
                [5, 1, 120, 1, 'RE', 40],
                [3, 2, 240, 0, 'HS', 80],
                [3, 1, 200, 0, 'HS', 80],
                [3, 1, 184, 0, 'HS', 80],
                [3, 1, 184, 0, 'HS', 80],
                [3, 1, 480, 1, 'HS', 112],
                [3, 1, 672, 1, 'HS', 112],
                [5, 2, 672, 1, 'HS', 160],
                [5, 1, 960, 1, 'HS', 160],
                [5, 1, 960, 1, 'HS', 160],
            ]

        base_channels = round_to_multiple_of(base_channels * width_multiplier, round_nearest)
        out_channels = round_to_multiple_of(out_channels * width_multiplier, round_nearest)

        for i in range(len(layer_setting)):
            # 缩放膨胀通道数
            layer_setting[i][2] = round_to_multiple_of(layer_setting[i][2], round_nearest)
            # 缩放输出通道数
            layer_setting[i][-1] = round_to_multiple_of(layer_setting[i][-1] * width_multiplier, round_nearest)

        self.first_stem, self.last_stem = make_stem(in_channels,
                                                    base_channels,
                                                    layer_setting[-1][-1],
                                                    out_channels,
                                                    conv_layer,
                                                    norm_layer,
                                                    act_layer)

        features = list()
        in_channels = base_channels
        for i, (kernel_size, stride, inner_planes, with_attention_2, non_linearity, out_channels) in enumerate(
                layer_setting):
            act_layer = relu_or_hswish(non_linearity)
            features.append(block_layer(in_channels,
                                        inner_planes,
                                        out_channels,
                                        stride=stride,
                                        kernel_size=kernel_size,
                                        with_attention=with_attention_2 and with_attention,
                                        reduction=reduction,
                                        attention_type=attention_type,
                                        conv_layer=conv_layer,
                                        norm_layer=norm_layer,
                                        act_layer=act_layer,
                                        sigmoid_type=sigmoid_type
                                        ))
            in_channels = out_channels
        self.add_module('features', nn.Sequential(*features))

        self._init_weights()

    def _init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.kaiming_uniform_(m.weight, mode="fan_out", nonlinearity="sigmoid")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def forward(self, x):
        x = self.first_stem(x)
        x = self.features(x)
        x = self.last_stem(x)
        return x


@registry.Backbone.register('MobileNetV3')
def build_mbv3_backbone(cfg):
    arch = cfg.MODEL.BACKBONE.ARCH
    in_channels = cfg.MODEL.BACKBONE.IN_PLANES
    norm_layer = get_norm(cfg)
    # compression
    width_multiplier = cfg.MODEL.COMPRESSION.WIDTH_MULTIPLIER
    round_nearest = cfg.MODEL.COMPRESSION.ROUND_NEAREST
    # attention
    with_attention = cfg.MODEL.ATTENTION.WITH_ATTENTION
    reduction = cfg.MODEL.ATTENTION.REDUCTION
    attention_type = cfg.MODEL.ATTENTION.ATTENTION_TYPE
    # conv
    conv_layer = get_conv(cfg)
    # act
    act_layer = get_act(cfg)
    sigmoid_type = cfg.MODEL.ACT.SIGMOID_TYPE

    base_channels, feature_dims, inner_dims, layer_setting = copy.deepcopy(arch_settings[arch])

    return MobileNetV3Backbone(
        in_channels=in_channels,
        base_channels=base_channels,
        out_channels=feature_dims,
        layer_setting=layer_setting,
        width_multiplier=width_multiplier,
        round_nearest=round_nearest,
        with_attention=with_attention,
        reduction=reduction,
        attention_type=attention_type,
        conv_layer=conv_layer,
        norm_layer=norm_layer,
        act_layer=act_layer,
        sigmoid_type=sigmoid_type
    )



# -*- coding: utf-8 -*-

"""
@date: 2020/4/27 下午8:25
@file: metrics.py
@author: zj
@description: 
"""

from thop import profile
from thop import clever_format


def compute_num_flops(model, data):
    macs, params = profile(model, inputs=(data,), verbose=False)
    # print(macs, params)

    GFlops = macs * 2.0 / pow(10, 9)
    # 假定每个参数使用32位浮点数保存，那么需要4个字节
    params_size = params * 4.0 / 1024 / 1024
    return GFlops, params_size


def topk_accuracy(output, target, top_k=(1,)):
    """
    计算前K个。N表示样本数，C表示类别数
    :param output: 大小为[N, C]，每行表示该样本计算得到的C个类别概率
    :param target: 大小为[N]，每行表示指定类别
    :param topk: tuple，计算前top-k的accuracy
    :return: list
    """
    assert len(output.shape) == 2 and output.shape[0] == target.shape[0]
    maxk = max(top_k)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, largest=True, sorted=True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in top_k:
        # correct_k = correct[:k].view(-1).float().sum(0)
        correct_k = correct[:k].reshape(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res



# -*- coding: utf-8 -*-

"""
@date: 2021/10/3 下午12:06
@file: color_jitter.py
@author: zj
@description: 
"""

import albumentations as A


class ColorJitter(object):
    """
    Randomly changes the brightness, contrast, and saturation of an image. Compared to ColorJitter from torchvision,
    this transform gives a little bit different results because Pillow (used in torchvision) and OpenCV (used in
    Albumentations) transform an image to HSV format by different formulas. Another difference - Pillow uses uint8
    overflow, but we use value saturation.

    Args:
        brightness (float or tuple of float (min, max)): How much to jitter brightness.
            brightness_factor is chosen uniformly from [max(0, 1 - brightness), 1 + brightness]
            or the given [min, max]. Should be non negative numbers.
        contrast (float or tuple of float (min, max)): How much to jitter contrast.
            contrast_factor is chosen uniformly from [max(0, 1 - contrast), 1 + contrast]
            or the given [min, max]. Should be non negative numbers.
        saturation (float or tuple of float (min, max)): How much to jitter saturation.
            saturation_factor is chosen uniformly from [max(0, 1 - saturation), 1 + saturation]
            or the given [min, max]. Should be non negative numbers.
        hue (float or tuple of float (min, max)): How much to jitter hue.
            hue_factor is chosen uniformly from [-hue, hue] or the given [min, max].
            Should have 0 <= hue <= 0.5 or -0.5 <= min <= max <= 0.5.
    """

    def __init__(self, brightness=0.2, contrast=0.2, saturation=0.2, hue=0.2, p=0.5):
        self.brightness = brightness
        self.contrast = contrast
        self.saturation = saturation
        self.hue = hue
        self.p = p

        self.t = A.ColorJitter(brightness=brightness, contrast=self.contrast, saturation=self.saturation, p=self.p)

    def __call__(self, image):
        return self.t(image=image)['image']

    def __repr__(self):
        return self.__class__.__name__ + '(brightness={0}, contrast={1}, saturation=(2), p={3})' \
            .format(self.brightness, self.contrast, self.saturation, self.hue, self.p)



# -*- coding: utf-8 -*-

"""
@date: 2021/2/23 下午8:22
@file: fashionmnist.py
@author: zj
@description: 
"""

from torch.utils.data import Dataset
import torchvision.datasets as datasets

from .evaluator.general_evaluator import GeneralEvaluator


class FashionMNIST(Dataset):

    def __init__(self, root, train=True, transform=None, target_transform=None, top_k=(1, 5)):
        self.data_set = datasets.FashionMNIST(root, train=train, transform=transform, target_transform=target_transform,
                                              download=True)
        self.classes = self.data_set.classes
        self.root = root
        self._update_evaluator(top_k)

    def __getitem__(self, index: int):
        return self.data_set.__getitem__(index)

    def __len__(self) -> int:
        return self.data_set.__len__()

    def _update_evaluator(self, top_k):
        self.evaluator = GeneralEvaluator(self.classes, top_k=top_k)

    def __repr__(self):
        return self.__class__.__name__ + ' (' + self.root + ')'



# -*- coding: utf-8 -*-

"""
@date: 2021/10/3 下午12:06
@file: color_jitter.py
@author: zj
@description: 
"""

import albumentations as A


class Normalize(object):
    """
    Normalization is applied by the formula: `img = (img - mean * max_pixel_value) / (std * max_pixel_value)`

    Args:
        mean (float, list of float): mean values
        std  (float, list of float): std values
        max_pixel_value (float): maximum possible pixel value

    Image types:
        uint8, float32
    """

    def __init__(self, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225), max_pixel_value=255.0, p=1.0):
        self.mean = mean
        self.std = std
        self.max_pixel_value = max_pixel_value
        self.p = p

        self.t = A.Normalize(mean=self.mean, std=self.std, max_pixel_value=self.max_pixel_value, p=self.p)

    def __call__(self, image):
        return self.t(image=image)['image']

    def __repr__(self):
        return self.__class__.__name__ + '(mean={0}, std={1}, max_pixel_value={2}, p={3})' \
            .format(self.mean, self.std, self.max_pixel_value, self.p)



# -*- coding: utf-8 -*-

"""
@date: 2020/11/21 下午3:15
@file: bottleneck.py
@author: zj
@description: 
"""
from abc import ABC

import torch.nn as nn


class ResNet3DBottleneck(nn.Module, ABC):
    expansion = 4

    def __init__(self,
                 in_planes,
                 out_planes,
                 spatial_stride=1,
                 temporal_stride=1,
                 down_sample=None,
                 inflate=False,
                 inflate_style='3x1x1',
                 groups=1,
                 base_width=64,
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None
                 ):
        """
        依次执行大小为Tx1x1、Tx3x3、1x1x1的卷积操作，
        如果进行下采样，那么使用第二个卷积层对输入空间尺寸进行减半操作
        分两种膨胀类型：3x1x1和3x3x3
        :param in_planes: 输入通道数
        :param out_planes: 输出通道数
        :param spatial_stride: 空间步长
        :param temporal_stride: 时间步长
        :param down_sample: 下采样
        :param inflate: 是否膨胀
        :param inflate_style: 膨胀类型，作用于Bottleneck
        :param groups: cardinality
        :param base_width: 基础宽度
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super(ResNet3DBottleneck, self).__init__()
        assert inflate_style in ("3x1x1", "3x3x3")

        if conv_layer is None:
            conv_layer = nn.Conv3d
        if norm_layer is None:
            norm_layer = nn.BatchNorm3d
        if act_layer is None:
            act_layer = nn.ReLU

        if inflate:
            if inflate_style == '3x1x1':
                conv1_kernel_size = (3, 1, 1)
                conv1_stride = (temporal_stride, 1, 1)
                conv1_padding = (1, 0, 0)
                conv2_kernel_size = (1, 3, 3)
                conv2_stride = (1, spatial_stride, spatial_stride)
                conv2_padding = (0, 1, 1)
            else:
                conv1_kernel_size = (1, 1, 1)
                conv1_stride = (1, 1, 1)
                conv1_padding = (0, 0, 0)
                conv2_kernel_size = (3, 3, 3)
                conv2_stride = (temporal_stride, spatial_stride, spatial_stride)
                conv2_padding = (1, 1, 1)
        else:
            conv1_kernel_size = (1, 1, 1)
            conv1_stride = (1, 1, 1)
            conv1_padding = (0, 0, 0)
            conv2_kernel_size = (1, 3, 3)
            conv2_stride = (1, spatial_stride, spatial_stride)
            conv2_padding = (0, 1, 1)

        self.down_sample = down_sample

        width = int(out_planes * (base_width / 64.)) * groups
        self.conv1 = conv_layer(in_planes, width, kernel_size=conv1_kernel_size,
                                stride=conv1_stride, padding=conv1_padding, bias=False)
        self.bn1 = norm_layer(width)

        self.conv2 = conv_layer(width, width, kernel_size=conv2_kernel_size,
                                stride=conv2_stride, padding=conv2_padding, bias=False,
                                groups=groups)
        self.bn2 = norm_layer(width)

        self.conv3 = conv_layer(width, out_planes * self.expansion, kernel_size=(1, 1, 1), stride=(1, 1, 1), bias=False)
        self.bn3 = norm_layer(out_planes * self.expansion)

        self.relu = act_layer(inplace=True)

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.down_sample is not None:
            identity = self.down_sample(x)

        out += identity
        out = self.relu(out)

        return out



# -*- coding: utf-8 -*-

"""
@date: 2020/12/30 下午4:28
@file: mnasnet_unit.py
@author: zj
@description:
"""
from abc import ABC

import torch.nn as nn

from zcls.model.attention_helper import make_attention_block

# Paper suggests 0.9997 momentum, for TensorFlow. Equivalent PyTorch momentum is
# 1.0 - tensorflow.
BN_MOMENTUM = 1 - 0.9997


class MNASNetUint(nn.Module, ABC):

    def __init__(self,
                 in_planes,
                 out_planes,
                 stride=1,
                 kernel_size=3,
                 expansion_rate=1,
                 with_attention=True,
                 reduction=4,
                 attention_type='SqueezeAndExcitationBlock2D',
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None
                 ):
        """
        整合了MBConv3(3x3/5x5/SE)、MBConv6(3x3/5x5/SE)、SepConv(k3x3)实现
        :param in_planes: 输入通道数
        :param out_planes: 输出通道数
        :param stride: 步长
        :param kernel_size: 卷积核大小
        :param expansion_rate: 膨胀因子，适用于mbv2的反向残差块
        :param with_attention: 是否使用注意力模块
        :param reduction: 衰减率
        :param attention_type: 注意力模块类型
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super(MNASNetUint, self).__init__()
        assert isinstance(expansion_rate, int) and expansion_rate >= 1

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU

        # if expansion_rate, use MBConv6(k3x3)/MBConv3(k5x5), else use SepConv(k3x3)
        self.with_expansion = expansion_rate > 1
        # if with attention, use MBConv3(k5x5)
        self.with_attention = with_attention

        inner_planes = in_planes * expansion_rate
        if self.with_expansion:
            self.expansion = nn.Sequential(
                conv_layer(in_planes, inner_planes, kernel_size=1, stride=1, padding=0, bias=False),
                norm_layer(inner_planes, momentum=BN_MOMENTUM),
                act_layer(inplace=True)
            )

        if kernel_size == 3:
            padding = 1
        elif kernel_size == 5:
            padding = 2
        else:
            padding = 0
        self.conv1 = conv_layer(inner_planes, inner_planes, kernel_size=kernel_size, stride=stride, padding=padding,
                                bias=False, groups=inner_planes)
        self.norm1 = norm_layer(inner_planes, momentum=BN_MOMENTUM)
        self.act = act_layer(inplace=True)

        self.conv2 = conv_layer(inner_planes, out_planes, kernel_size=1, stride=1, padding=0, bias=False)
        self.norm2 = norm_layer(out_planes, momentum=BN_MOMENTUM)

        if self.with_attention:
            self.attention = make_attention_block(inner_planes, reduction, attention_type)

        self.apply_residual = (in_planes == out_planes and stride == 1)

    def forward(self, x):
        identity = x

        if self.with_expansion:
            x = self.expansion(x)

        out = self.conv1(x)
        out = self.norm1(out)
        out = self.act(out)

        if self.with_attention:
            out = self.attention(out)

        out = self.conv2(out)
        out = self.norm2(out)

        if self.apply_residual:
            out = out + identity
        # 参考Torchvision.
        # Linear pointwise. Note that there's no activation.
        # out = self.act(out)

        return out



# -*- coding: utf-8 -*-

"""
@date: 2021/7/28 下午5:50
@file: ddb_util.py
@author: zj
@description: 
"""

from .dbb_transforms import transI_fusebn, transII_addbranch, transIII_1x1_kxk, \
    transVI_multiscale, transV_avg, transVI_multiscale

from .diverse_branch_block import DiverseBranchBlock


def get_equivalent_kernel_bias(module):
    assert isinstance(module, DiverseBranchBlock)

    k_origin, b_origin = transI_fusebn(module.dbb_origin.conv.weight, module.dbb_origin.bn)

    if hasattr(module, 'dbb_1x1'):
        k_1x1, b_1x1 = transI_fusebn(module.dbb_1x1.conv.weight, module.dbb_1x1.bn)
        k_1x1 = transVI_multiscale(k_1x1, module.kernel_size)
    else:
        k_1x1, b_1x1 = 0, 0

    if hasattr(module.dbb_1x1_kxk, 'idconv1'):
        k_1x1_kxk_first = module.dbb_1x1_kxk.idconv1.get_actual_kernel()
    else:
        k_1x1_kxk_first = module.dbb_1x1_kxk.conv1.weight
    k_1x1_kxk_first, b_1x1_kxk_first = transI_fusebn(k_1x1_kxk_first, module.dbb_1x1_kxk.bn1)
    k_1x1_kxk_second, b_1x1_kxk_second = transI_fusebn(module.dbb_1x1_kxk.conv2.weight, module.dbb_1x1_kxk.bn2)
    k_1x1_kxk_merged, b_1x1_kxk_merged = transIII_1x1_kxk(k_1x1_kxk_first, b_1x1_kxk_first, k_1x1_kxk_second,
                                                          b_1x1_kxk_second, groups=module.groups)

    k_avg = transV_avg(module.out_channels, module.kernel_size, module.groups)
    k_1x1_avg_second, b_1x1_avg_second = transI_fusebn(k_avg.to(module.dbb_avg.avgbn.weight.device),
                                                       module.dbb_avg.avgbn)
    if hasattr(module.dbb_avg, 'conv'):
        k_1x1_avg_first, b_1x1_avg_first = transI_fusebn(module.dbb_avg.conv.weight, module.dbb_avg.bn)
        k_1x1_avg_merged, b_1x1_avg_merged = transIII_1x1_kxk(k_1x1_avg_first, b_1x1_avg_first, k_1x1_avg_second,
                                                              b_1x1_avg_second, groups=module.groups)
    else:
        k_1x1_avg_merged, b_1x1_avg_merged = k_1x1_avg_second, b_1x1_avg_second

    return transII_addbranch((k_origin, k_1x1, k_1x1_kxk_merged, k_1x1_avg_merged),
                             (b_origin, b_1x1, b_1x1_kxk_merged, b_1x1_avg_merged))



# -*- coding: utf-8 -*-

"""
@date: 2021/7/23 下午9:27
@file: square_pad.py
@author: zj
@description: 
"""

import cv2
import albumentations as A


class SquarePad(object):
    """
    By filling the shorter edges, the size of the image becomes square

    Args:
        padding_position (Union[str, PositionType]): Position of the image. should be PositionType.CENTER or
            PositionType.TOP_LEFT or PositionType.TOP_RIGHT or PositionType.BOTTOM_LEFT or PositionType.BOTTOM_RIGHT.
            Default: PositionType.CENTER.
        padding_mode (OpenCV flag): OpenCV border mode.
        fill (int, float, list of int, list of float): padding value if border_mode is cv2.BORDER_CONSTANT.
        p (float): probability of applying the transform. Default: 1.0.
    """

    def __init__(self, padding_position=A.PadIfNeeded.PositionType.CENTER, padding_mode=cv2.BORDER_CONSTANT,
                 fill=0, p=1.0):
        self.fill = fill
        self.padding_position = padding_position
        self.padding_mode = padding_mode
        self.p = p

        self.t = A.PadIfNeeded(min_width=224, min_height=224, position=self.padding_position,
                               border_mode=self.padding_mode, value=self.fill, p=self.p)

    def __call__(self, image):
        h, w = image.shape[:2]

        min_size = h if h > w else w
        self.t.min_width = min_size
        self.t.min_height = min_size
        return self.t(image=image)['image']

    def __repr__(self):
        return self.__class__.__name__ + '(padding_position={0}, padding_mode={1}, fill={2}, p={3})'. \
            format(self.padding_position, self.padding_mode, self.fill, self.p)



# -*- coding: utf-8 -*-

"""
@date: 2021/10/3 上午11:27
@file: random_crop.py
@author: zj
@description: 
"""

import albumentations as A
from typing import Sequence


class CenterCrop(object):
    """
    Crop the central part of the input.

    Args:
        size (sequence): Desired output size.
        p (float): probability of applying the transform. Default: 1.

    Image types:
        uint8, float32

    Note:
        It is recommended to use uint8 images as input.
        Otherwise the operation will require internal conversion
        float32 -> uint8 -> float32 that causes worse performance.
    """

    def __init__(self, size, p=1.0):
        if not isinstance(size, Sequence):
            raise TypeError("Size should be int or sequence. Got {}".format(type(size)))
        if isinstance(size, Sequence) and len(size) not in (2,):
            raise ValueError("If size is a sequence, it should have 2 values")
        self.size = size
        self.p = p

        new_h, new_w = self.size
        self.t = A.CenterCrop(width=new_w, height=new_h, p=1.0)

    def __call__(self, image):
        return self.t(image=image)['image']

    def __repr__(self):
        return self.__class__.__name__ + '(size={0}, p={1})'. \
            format(self.size, self.p)



# -*- coding: utf-8 -*-

"""
@date: 2020/8/22 下午9:40
@file: trainer.py
@author: zj
@description: 
"""

import torchvision.transforms.transforms as transforms
from torchvision.transforms.autoaugment import AutoAugmentPolicy

from . import realization


def parse_transform(cfg, is_train=True):
    methods = cfg.TRANSFORM.TRAIN_METHODS if is_train else cfg.TRANSFORM.TEST_METHODS
    assert isinstance(methods, tuple)

    keys = realization.__all__
    transforms_dict = realization.__dict__
    aug_list = list()
    for method in methods:
        if method in keys:
            transform = transforms_dict[method]
        else:
            raise ValueError(f'f{method} does not exists')

        if method == 'AutoAugment':
            policy, p = cfg.TRANSFORM.AUTOAUGMENT
            if policy == 'imagenet':
                policy = AutoAugmentPolicy.IMAGENET
            elif policy == 'cifar10':
                policy = AutoAugmentPolicy.CIFAR10
            elif policy == 'svhn':
                policy = AutoAugmentPolicy.SVHN
            else:
                raise ValueError(f'{policy} does not supports')
            aug_list.append(transform(policy=policy, p=p))
        elif method == 'CoarseDropout':
            max_holes, max_height, max_width, min_holes, min_height, min_width, fill_value, p \
                = cfg.TRANSFORM.COARSE_DROPOUT
            aug_list.append(transform(max_holes=max_holes, max_height=max_height, max_width=max_width,
                                      min_holes=min_holes, min_height=min_height, min_width=min_width,
                                      fill_value=fill_value, p=p))
        elif method == 'CenterCrop':
            if is_train:
                size, p = cfg.TRANSFORM.TRAIN_CENTER_CROP
            else:
                size, p = cfg.TRANSFORM.TEST_CENTER_CROP
            aug_list.append(transform(size, p=p))
        elif method == 'ColorJitter':
            brightness, contrast, saturation, hue, p = cfg.TRANSFORM.COLOR_JITTER
            aug_list.append(transform(brightness=brightness, contrast=contrast, saturation=saturation, hue=hue, p=p))
        elif method == 'RandomCrop':
            size, p = cfg.TRANSFORM.RANDOM_CROP
            aug_list.append(transform(size, p=p))
        elif method == 'HorizontalFlip':
            p = cfg.TRANSFORM.HORIZONTAL_FLIP
            aug_list.append(transform(p=p))
        elif method == 'VerticalFlip':
            p = cfg.TRANSFORM.VERTICAL_FLIP
            aug_list.append(transform(p=p))
        elif method == 'Resize':
            if is_train:
                size, interpolation, p = cfg.TRANSFORM.TRAIN_RESIZE
            else:
                size, interpolation, p = cfg.TRANSFORM.TEST_RESIZE
            aug_list.append(transform(size, interpolation=interpolation, p=p))
        elif method == 'Rotate':
            limit, interpolation, border_mode, value, p = cfg.TRANSFORM.ROTATE
            aug_list.append(transform(limit, interpolation=interpolation,
                                      border_mode=border_mode, value=value, p=p))
        elif method == 'SquarePad':
            padding_position, padding_mode, fill, p = cfg.TRANSFORM.SQUARE_PAD
            aug_list.append(transform(padding_position=padding_position, padding_mode=padding_mode, fill=fill, p=p))
        elif method == 'Normalize':
            mean, std, max_pixel_value, p = cfg.TRANSFORM.NORMALIZE
            aug_list.append(transform(mean=mean, std=std, max_pixel_value=max_pixel_value, p=p))
        elif method == 'ToTensor':
            p = cfg.TRANSFORM.TO_TENSOR
            aug_list.append(transform(p=p))
        else:
            raise ValueError(f'{method} does not exists')

    return transforms.Compose(aug_list)


def parse_target_transform():
    return None


def build_transform(cfg, is_train=True):
    return parse_transform(cfg, is_train), parse_target_transform()



# -*- coding: utf-8 -*-

"""
@date: 2021/2/19 下午6:51
@file: torchvision_mobilenetv2.py
@author: zj
@description: 
"""

from abc import ABC
import torch.nn as nn
from torchvision.models import mobilenet_v3_large, mobilenet_v3_small

from zcls.config.key_word import KEY_OUTPUT
from zcls.model import registry
from zcls.model.norm_helper import freezing_bn, get_norm


class TorchvisionMobileNetV3(nn.Module, ABC):

    def __init__(self,
                 arch='mobilenet_v3_large',
                 width_multiplier=1.0,
                 num_classes=1000,
                 torchvision_pretrained=False,
                 pretrained_num_classes=1000,
                 fix_bn=False,
                 partial_bn=False,
                 norm_layer=None):
        super(TorchvisionMobileNetV3, self).__init__()
        self.fix_bn = fix_bn
        self.partial_bn = partial_bn

        if arch == 'mobilenet_v3_large':
            self.model = mobilenet_v3_large(pretrained=torchvision_pretrained,
                                            _width_mult=width_multiplier,
                                            norm_layer=norm_layer,
                                            num_classes=pretrained_num_classes)
        elif arch == 'mobilenet_v3_small':
            self.model = mobilenet_v3_small(pretrained=torchvision_pretrained,
                                            _width_mult=width_multiplier,
                                            norm_layer=norm_layer,
                                            num_classes=pretrained_num_classes)
        else:
            raise ValueError(f"{arch} doesn't exists")

        self.init_weights(num_classes, pretrained_num_classes)

    def init_weights(self, num_classes, pretrained_num_classes):
        if num_classes != pretrained_num_classes:
            fc = self.model.classifier[1]
            fc_features = fc.in_features
            fc = nn.Linear(fc_features, num_classes)

            nn.init.normal_(fc.weight, 0, 0.01)
            nn.init.zeros_(fc.bias)

            self.model.classifier[1] = fc

    def train(self, mode: bool = True):
        super(TorchvisionMobileNetV3, self).train(mode=mode)

        if mode and (self.partial_bn or self.fix_bn):
            freezing_bn(self, partial_bn=self.partial_bn)

        return self

    def forward(self, x):
        x = self.model(x)

        return {KEY_OUTPUT: x}


@registry.RECOGNIZER.register('TorchvisionMobileNetV3')
def build_torchvision_mbv3(cfg):
    arch = cfg.MODEL.BACKBONE.ARCH
    torchvision_pretrained = cfg.MODEL.RECOGNIZER.TORCHVISION_PRETRAINED
    pretrained_num_classes = cfg.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES
    fix_bn = cfg.MODEL.NORM.FIX_BN
    partial_bn = cfg.MODEL.NORM.PARTIAL_BN
    num_classes = cfg.MODEL.HEAD.NUM_CLASSES
    norm_layer = get_norm(cfg)
    width_multiplier = cfg.MODEL.COMPRESSION.WIDTH_MULTIPLIER

    return TorchvisionMobileNetV3(
        arch=arch,
        width_multiplier=width_multiplier,
        num_classes=num_classes,
        torchvision_pretrained=torchvision_pretrained,
        pretrained_num_classes=pretrained_num_classes,
        fix_bn=fix_bn,
        partial_bn=partial_bn,
        norm_layer=norm_layer
    )



# -*- coding: utf-8 -*-

"""
@date: 2020/12/3 下午8:52
@file: mobilenetv2_backbone.py
@author: zj
@description: 
"""

from abc import ABC
import copy
import torch.nn as nn

from zcls.model import registry
from zcls.model.init_helper import init_weights
from zcls.model.conv_helper import get_conv
from zcls.model.norm_helper import get_norm
from zcls.model.act_helper import get_act
from .shufflenetv1_unit import ShuffleNetV1Unit

arch_settings = {
    # block_layer, groups, base_channel, stage_channels, layer_blocks, width_multiplier
    'shufflenetv1_3g2x': (ShuffleNetV1Unit, 3, 48, [240, 480, 960], (4, 8, 4), 2.0),
    'shufflenetv1_3g1_5x': (ShuffleNetV1Unit, 3, 24, [240, 480, 960], (4, 8, 4), 1.5),
    'shufflenetv1_3g1x': (ShuffleNetV1Unit, 3, 24, [240, 480, 960], (4, 8, 4), 1.0),
    'shufflenetv1_3g0_5x': (ShuffleNetV1Unit, 3, 12, [240, 480, 960], (4, 8, 4), 0.5),
    'shufflenetv1_8g2x': (ShuffleNetV1Unit, 8, 48, [384, 768, 1536], (4, 8, 4), 2.0),
    'shufflenetv1_8g1_5x': (ShuffleNetV1Unit, 8, 24, [384, 768, 1536], (4, 8, 4), 1.5),
    'shufflenetv1_8g1x': (ShuffleNetV1Unit, 8, 24, [384, 768, 1536], (4, 8, 4), 1.0),
    'shufflenetv1_8g0_5x': (ShuffleNetV1Unit, 8, 16, [384, 768, 1536], (4, 8, 4), 0.5),
}


def make_stage(in_channels,
               out_channels,
               groups,
               block_num,
               with_downsample,
               with_groups,
               block_layer,
               conv_layer,
               norm_layer,
               act_layer,
               ):
    """
    :param in_channels: 输入通道数
    :param out_channels: 输出通道数
    :param groups: 分组数
    :param block_num: 块个数
    :param with_downsample: 是否执行空间下采样
    :param with_groups: 是否对第一个1x1逐点卷积执行分组操作
    :param block_layer: 块类型
    :param conv_layer: 卷积层类型
    :param norm_layer: 归一化层类型
    :param act_layer: 激活层类型
    :return:
    """
    with_groups = with_groups if isinstance(with_groups, tuple) else [with_groups] * block_num
    assert len(with_groups) == block_num
    stride = 2 if with_downsample else 1
    down_sample = nn.AvgPool2d(3, stride=stride, padding=1) if stride == 2 else None

    blocks = list()
    blocks.append(block_layer(
        in_channels, out_channels, groups, stride, down_sample, with_groups[0], conv_layer, norm_layer, act_layer))
    in_channels = out_channels

    stride = 1
    down_sample = None
    for i in range(1, block_num):
        blocks.append(block_layer(
            in_channels, out_channels, groups, stride, down_sample, True, conv_layer, norm_layer, act_layer))
    return nn.Sequential(*blocks)


def make_stem(in_planes,
              base_planes,
              conv_layer,
              norm_layer,
              act_layer
              ):
    return nn.Sequential(
        conv_layer(in_planes, base_planes, kernel_size=3, stride=2, padding=1, bias=False),
        norm_layer(base_planes),
        act_layer(inplace=True),
        nn.MaxPool2d(3, stride=2, padding=1)
    )


class ShuffleNetV1Backbone(nn.Module, ABC):

    def __init__(self,
                 in_channels=3,
                 base_channels=24,
                 groups=8,
                 stage_channels=(384, 768, 1536),
                 stage_blocks=(4, 8, 4),
                 downsamples=(1, 1, 1),
                 with_groups=(0, 1, 1),
                 block_layer=None,
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None,
                 ):
        """
        :param in_channels: 输入通道数
        :param base_channels: 第一个卷积层通道数
        :param groups: 分组数
        :param stage_channels: 每个阶段通道数
        :param stage_blocks: 每个阶段块个数
        :param downsamples: 是否执行空间下采样
        :param with_groups: 是否对第一个1x1逐点卷积执行分组操作
        :param block_layer: 块类型
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super(ShuffleNetV1Backbone, self).__init__()

        if block_layer is None:
            block_layer = ShuffleNetV1Unit
        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU

        self.stem = make_stem(in_channels,
                              base_channels,
                              conv_layer,
                              norm_layer,
                              act_layer
                              )

        in_channels = base_channels
        for i in range(len(stage_blocks)):
            res_layer = make_stage(in_channels,
                                   stage_channels[i],
                                   groups,
                                   stage_blocks[i],
                                   downsamples[i],
                                   with_groups[i],
                                   block_layer,
                                   conv_layer,
                                   norm_layer,
                                   act_layer
                                   )
            in_channels = stage_channels[i]
            stage_name = f'stage{i + 1}'
            self.add_module(stage_name, res_layer)
        self.stage_num = len(stage_blocks)

        init_weights(self.modules())

    def forward(self, x):
        x = self.stem(x)

        for i in range(self.stage_num):
            stage = self.__getattr__(f'stage{i + 1}')
            x = stage(x)

        return x


@registry.Backbone.register('ShuffleNetV1')
def build_sfv1_backbone(cfg):
    arch = cfg.MODEL.BACKBONE.ARCH
    in_channels = cfg.MODEL.BACKBONE.IN_PLANES
    downsamples = cfg.MODEL.BACKBONE.DOWNSAMPLES
    with_groups = cfg.MODEL.BACKBONE.WITH_GROUPS
    conv_layer = get_conv(cfg)
    norm_layer = get_norm(cfg)
    act_layer = get_act(cfg)

    block_layer, groups, base_channels, stage_channels, layer_blocks, width_multiplier = copy.deepcopy(
        arch_settings[arch])

    for i in range(len(stage_channels)):
        stage_channels[i] = int(stage_channels[i] * width_multiplier)

    return ShuffleNetV1Backbone(
        in_channels=in_channels,
        base_channels=base_channels,
        groups=groups,
        stage_channels=stage_channels,
        stage_blocks=layer_blocks,
        downsamples=downsamples,
        with_groups=with_groups,
        block_layer=block_layer,
        conv_layer=conv_layer,
        norm_layer=norm_layer,
        act_layer=act_layer,
    )



# -*- coding: utf-8 -*-

"""
@date: 2021/9/23 下午10:06
@file: general_dataset_v2.py
@author: zj
@description: 
"""

import os
import json

from PIL import Image

from torch.utils.data import Dataset
from torchvision.datasets.folder import default_loader

from .evaluator.general_evaluator import GeneralEvaluator


class GeneralDatasetV2(Dataset):

    def __init__(self, root, transform=None, target_transform=None, top_k=(1, 5)):
        assert os.path.isfile(root)
        with open(root, 'r') as f:
            data_dict = json.load(f)

        self.classes = list(data_dict.keys())
        self.total_img_list = list()
        self.total_label_list = list()
        for key in self.classes:
            img_list = data_dict[key]

            label = self.classes.index(key)
            for img_path in img_list:
                assert os.path.isfile(img_path), img_path

                self.total_img_list.append(img_path)
                self.total_label_list.append(label)

        self.root = root
        self.transform = transform
        self.target_transform = target_transform
        self._update_evaluator(top_k)

    def __getitem__(self, index: int):
        img_path = self.total_img_list[index]
        target = self.total_label_list[index]

        image = default_loader(img_path)
        if self.transform is not None:
            image = self.transform(image)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return image, target

    def __len__(self) -> int:
        return len(self.total_img_list)

    def _update_evaluator(self, top_k):
        self.evaluator = GeneralEvaluator(self.classes, top_k=top_k)

    def get_classes(self):
        return self.classes

    def __repr__(self):
        return self.__class__.__name__ + ' (' + self.root + ')'



# -*- coding: utf-8 -*-

"""
@date: 2021/7/26 下午10:10
@file: cutmix.py
@author: zj
@description:
refer to [ clovaai/CutMix-PyTorch](https://github.com/clovaai/CutMix-PyTorch)
"""

import torch

import numpy as np

from zcls.config.key_word import KEY_LOSS


def rand_bbox(size, lam):
    W = size[2]
    H = size[3]
    cut_rat = np.sqrt(1. - lam)
    cut_w = np.int(W * cut_rat)
    cut_h = np.int(H * cut_rat)

    # uniform
    cx = np.random.randint(W)
    cy = np.random.randint(H)

    bbx1 = np.clip(cx - cut_w // 2, 0, W)
    bby1 = np.clip(cy - cut_h // 2, 0, H)
    bbx2 = np.clip(cx + cut_w // 2, 0, W)
    bby2 = np.clip(cy + cut_h // 2, 0, H)

    return bbx1, bby1, bbx2, bby2


def cutmix_data(images, targets, alpha=1.0, device=torch.device('cpu')):
    '''
    Returns mixed inputs, pairs of targets, and lambda
    '''
    if alpha > 0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1

    batch_size = images.size()[0]
    rand_index = torch.randperm(batch_size).to(device)

    targets_a = targets
    targets_b = targets[rand_index]

    bbx1, bby1, bbx2, bby2 = rand_bbox(images.size(), lam)
    images[:, :, bbx1:bbx2, bby1:bby2] = images[rand_index, :, bbx1:bbx2, bby1:bby2]

    # adjust lambda to exactly match pixel ratio
    lam = 1 - ((bbx2 - bbx1) * (bby2 - bby1) / (images.size()[-1] * images.size()[-2]))

    return images, targets_a, targets_b, lam


def cutmix_criterion(criterion, output_dict: dict, targets_a: torch.Tensor, targets_b: torch.Tensor, lam):
    loss_a = criterion(output_dict, targets_a)[KEY_LOSS]
    loss_b = criterion(output_dict, targets_b)[KEY_LOSS]

    total_loss = lam * loss_a + (1 - lam) * loss_b

    return {KEY_LOSS: total_loss}


def cutmix_evaluate(evaluator, output_dict, targets_a, targets_b, lam):
    acc_dict_a = evaluator.evaluate_train(output_dict, targets_a)
    acc_dict_b = evaluator.evaluate_train(output_dict, targets_b)

    total_acc_dict = dict()

    for (a_key, a_value), (b_key, b_value) in zip(acc_dict_a.items(), acc_dict_b.items()):
        assert a_key == b_key
        total_acc_dict[a_key] = lam * a_value + (1 - lam) * b_value

    return total_acc_dict



# -*- coding: utf-8 -*-

"""
@date: 2020/12/24 下午7:38
@file: shufflenetv1.py
@author: zj
@description: 
"""

import torch.nn as nn
from torchvision.models.utils import load_state_dict_from_url

from zcls.model import registry
from ..base_recognizer import BaseRecognizer


@registry.RECOGNIZER.register('MobileNetV3')
class MobileNetV3(BaseRecognizer):

    def __init__(self, cfg):
        super().__init__(cfg)

    def init_weights(self, pretrained, pretrained_num_classes, num_classes):
        # Using super class method to load pretraining weights
        super(MobileNetV3, self).init_weights(pretrained, pretrained_num_classes, pretrained_num_classes)
        if num_classes != pretrained_num_classes:
            in_channels = self.head.conv2.in_channels
            conv2 = nn.Conv2d(in_channels, num_classes, kernel_size=1, stride=1, padding=0, bias=True)

            nn.init.kaiming_normal_(conv2.weight, mode="fan_out", nonlinearity="relu")
            nn.init.zeros_(conv2.bias)

            self.head.conv2 = conv2



# -*- coding: utf-8 -*-

"""
@date: 2020/11/25 下午8:44
@file: group_norm.py
@author: zj
@description: 
"""

import torch
import torch.nn as nn


class GroupNorm(nn.Module):
    """
    refer to [group_normalization/group_norm.py](https://github.com/taokong/group_normalization/blob/master/group_norm.py)
    """

    def __init__(self, num_groups: int = 4, num_channels: int = 32, eps: float = 1e-5, affine: bool = True):
        super(GroupNorm, self).__init__()
        self.num_groups = num_groups
        self.num_channels = num_channels
        self.eps = eps
        self.affine = affine

        if affine:
            self.weight = torch.ones(num_channels, 1, 1)
            self.bias = torch.zeros(num_channels, 1, 1)

    def forward(self, inputs):
        shape = inputs.size()
        N, C = shape[:2]

        x = inputs.view(N, self.num_groups, -1)

        mean = x.mean(dim=2, keepdim=True)
        std = x.std(dim=2, keepdim=True)

        x = (x - mean) / (std + self.eps)
        x = x.view(shape)

        return x * self.weight + self.bias



# -*- coding: utf-8 -*-

"""
@date: 2020/12/16 下午7:52
@file: global_context_block.py
@author: zj
@description: 
"""

import torch
import torch.nn as nn


class _GlobalContextBlockND(nn.Module):
    """
    refer to [context_block.py](https://github.com/xvjiarui/GCNet/blob/master/mmdet/ops/gcb/context_block.py)
    """

    def __init__(self,
                 in_channels,
                 reduction=16,
                 dimension=2
                 ):
        """
        :param in_channels:
        :param reduction:
        :param dimension:
        """
        super(_GlobalContextBlockND, self).__init__()
        assert dimension in [1, 2, 3]

        self.in_channels = in_channels
        self.reduction = reduction
        self.dimension = dimension

        if self.dimension == 1:
            self.conv_layer = nn.Conv1d
        elif self.dimension == 2:
            self.conv_layer = nn.Conv2d
        else:
            self.conv_layer = nn.Conv3d

        self._construct_gc()
        self.init_weights()

    def _construct_gc(self):
        self.w_k = self.conv_layer(self.in_channels, 1, kernel_size=1, stride=1, padding=0)

        reduction_channel = self.in_channels // self.reduction
        self.w_v1 = self.conv_layer(self.in_channels, reduction_channel, kernel_size=1, stride=1, padding=0)
        self.ln = nn.LayerNorm([reduction_channel, *([1] * self.dimension)])
        self.relu = nn.ReLU(inplace=True)
        self.w_v2 = self.conv_layer(reduction_channel, self.in_channels, kernel_size=1, stride=1, padding=0)

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')

    def forward(self, x):
        """
        :param x: (N, C, **)
        """
        self._check_input_dim(x)

        identity = x
        N, C = x.shape[:2]

        # (N, C, **) -> (N, C, D_j) -> (N, 1, C, D_j)
        input_x = x.view(N, C, -1).unsqueeze(1)
        # (N, C, **) -> (N, 1, **) -> (N, 1, D_j)
        context_mask = self.w_k(x).view(N, 1, -1)
        # (N, 1, D_j) -> (N, 1, D_j, 1)
        context_mask = nn.functional.softmax(context_mask, dim=2).unsqueeze(-1)
        # (N, 1, C, D_j) * (N, 1, D_j, 1) -> (N, 1, C, 1) -> (N, C) -> (N, C, **)
        context = torch.matmul(input_x, context_mask).reshape(N, C).reshape(N, C, *([1] * self.dimension))

        out = self.w_v1(context)
        out = self.ln(out)
        out = self.relu(out)
        transform = self.w_v2(out)

        z = transform + identity
        return z

    def _check_input_dim(self, input):
        raise NotImplementedError


class GlobalContextBlock1D(_GlobalContextBlockND):

    def __init__(self, in_channels, reduction=16, dimension=1):
        super().__init__(in_channels, reduction, dimension)

    def _check_input_dim(self, input):
        if input.dim() != 2 and input.dim() != 3:
            raise ValueError('expected 2D or 3D input (got {}D input)'
                             .format(input.dim()))


class GlobalContextBlock2D(_GlobalContextBlockND):

    def __init__(self, in_channels, reduction=16, dimension=2):
        super().__init__(in_channels, reduction, dimension)

    def _check_input_dim(self, input):
        if input.dim() != 4:
            raise ValueError('expected 4D input (got {}D input)'
                             .format(input.dim()))


class GlobalContextBlock3D(_GlobalContextBlockND):

    def __init__(self, in_channels, reduction=16, dimension=3):
        super().__init__(in_channels, reduction, dimension)

    def _check_input_dim(self, input):
        if input.dim() != 5:
            raise ValueError('expected 5D input (got {}D input)'
                             .format(input.dim()))



# -*- coding: utf-8 -*-

"""
@date: 2020/11/10 下午5:02
@file: cifar.py
@author: zj
@description: 
"""

from torch.utils.data import Dataset
from torchvision.datasets import CIFAR10, CIFAR100

from .evaluator.general_evaluator import GeneralEvaluator


class CIFAR(Dataset):

    def __init__(self, root, train=True, transform=None, target_transform=None, top_k=(1, 5), is_cifar100=True):
        if is_cifar100:
            self.data_set = CIFAR100(root, train=train, transform=transform, target_transform=target_transform,
                                     download=True)
        else:
            self.data_set = CIFAR10(root, train=train, transform=transform, target_transform=target_transform,
                                    download=True)
        self.classes = self.data_set.classes
        self.root = root
        self._update_evaluator(top_k)

    def __getitem__(self, index: int):
        return self.data_set.__getitem__(index)

    def __len__(self) -> int:
        return self.data_set.__len__()

    def _update_evaluator(self, top_k):
        self.evaluator = GeneralEvaluator(self.classes, top_k=top_k)

    def __repr__(self):
        return self.__class__.__name__ + ' (' + self.root + ')'



# -*- coding: utf-8 -*-

"""
@date: 2021/2/23 下午8:22
@file: imagenet.py
@author: zj
@description: 
"""

from torch.utils.data import Dataset
import torchvision.datasets as datasets

from .evaluator.general_evaluator import GeneralEvaluator


class ImageNet(Dataset):

    def __init__(self, root, train=True, transform=None, target_transform=None, top_k=(1, 5)):
        split = 'train' if train else 'val'
        # using torchvision ImageNet to get classes
        self.data_set = datasets.ImageNet(root, split=split, transform=transform, target_transform=target_transform)
        self.classes = list()
        for class_tuple in self.data_set.classes:
            self.classes.append(','.join(class_tuple))
        self.root = root
        # create evaluator
        self._update_evaluator(top_k)

    def __getitem__(self, index: int):
        return self.data_set.__getitem__(index)

    def __len__(self) -> int:
        return len(self.data_set)

    def _update_evaluator(self, top_k):
        self.evaluator = GeneralEvaluator(self.classes, top_k=top_k)

    def __repr__(self):
        return self.__class__.__name__ + ' (' + self.root + ')'



# -*- coding: utf-8 -*-

"""
@date: 2021/1/1 下午6:51
@file: rmsprop.py
@author: zj
@description: 
"""

import torch.optim as optim

from .. import registry


@registry.OPTIMIZERS.register('RMSProp')
def build_rmsprop(cfg, groups):
    lr = cfg.OPTIMIZER.LR
    weight_decay = cfg.OPTIMIZER.WEIGHT_DECAY.DECAY

    momentum = cfg.OPTIMIZER.MOMENTUM

    return optim.RMSprop(groups, lr=lr, momentum=momentum, weight_decay=weight_decay)



# -*- coding: utf-8 -*-

"""
@date: 2021/7/26 上午10:45
@file: mixup.py
@author: zj
@description:
refer to [facebookresearch/mixup-cifar10](https://github.com/facebookresearch/mixup-cifar10)
"""

import torch

import numpy as np

from zcls.config.key_word import KEY_LOSS


def mixup_data(images, targets, alpha=1.0, device=torch.device('cpu')):
    '''
    Returns mixed inputs, pairs of targets, and lambda
    '''
    if alpha > 0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1

    batch_size = images.size()[0]
    index = torch.randperm(batch_size).to(device)

    mixed_x = lam * images + (1 - lam) * images[index, :]
    targets_a, targets_b = targets, targets[index]
    return mixed_x, targets_a, targets_b, lam


def mixup_criterion(criterion, output_dict: dict, targets_a: torch.Tensor, targets_b: torch.Tensor, lam):
    loss_a = criterion(output_dict, targets_a)[KEY_LOSS]
    loss_b = criterion(output_dict, targets_b)[KEY_LOSS]

    total_loss = lam * loss_a + (1 - lam) * loss_b

    return {KEY_LOSS: total_loss}


def mixup_evaluate(evaluator, output_dict, targets_a, targets_b, lam):
    acc_dict_a = evaluator.evaluate_train(output_dict, targets_a)
    acc_dict_b = evaluator.evaluate_train(output_dict, targets_b)

    total_acc_dict = dict()

    for (a_key, a_value), (b_key, b_value) in zip(acc_dict_a.items(), acc_dict_b.items()):
        assert a_key == b_key
        total_acc_dict[a_key] = lam * a_value + (1 - lam) * b_value

    return total_acc_dict



# -*- coding: utf-8 -*-

"""
@date: 2021/2/20 上午10:28
@file: torchvision_resnet.py
@author: zj
@description: 
"""

from abc import ABC

import torch.nn as nn
from torch.nn.modules.module import T
from torchvision.models.resnet import resnet18, resnet34, resnet50, resnet101, resnet152, resnext50_32x4d, \
    resnext101_32x8d

from zcls.config.key_word import KEY_OUTPUT
from zcls.model import registry
from zcls.model.norm_helper import freezing_bn


class TorchvisionResNet(nn.Module, ABC):

    def __init__(self,
                 arch="resnet18",
                 num_classes=1000,
                 torchvision_pretrained=False,
                 pretrained_num_classes=1000,
                 fix_bn=False,
                 partial_bn=False,
                 zero_init_residual=False):
        super(TorchvisionResNet, self).__init__()

        self.num_classes = num_classes
        self.fix_bn = fix_bn
        self.partial_bn = partial_bn

        if arch == 'resnet18':
            self.model = resnet18(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes,
                                  zero_init_residual=zero_init_residual)
        elif arch == 'resnet34':
            self.model = resnet34(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes,
                                  zero_init_residual=zero_init_residual)
        elif arch == 'resnet50':
            self.model = resnet50(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes,
                                  zero_init_residual=zero_init_residual)
        elif arch == 'resnet101':
            self.model = resnet101(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes,
                                   zero_init_residual=zero_init_residual)
        elif arch == 'resnet152':
            self.model = resnet152(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes,
                                   zero_init_residual=zero_init_residual)
        elif arch == 'resnext50_32x4d':
            self.model = resnext50_32x4d(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes,
                                         zero_init_residual=zero_init_residual)
        elif arch == 'resnext101_32x8d':
            self.model = resnext101_32x8d(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes,
                                          zero_init_residual=zero_init_residual)
        else:
            raise ValueError('no such value')

        self.init_weights(num_classes, pretrained_num_classes)

    def init_weights(self, num_classes, pretrained_num_classes):
        if num_classes != pretrained_num_classes:
            fc = self.model.fc
            fc_features = fc.in_features
            self.model.fc = nn.Linear(fc_features, num_classes)

            nn.init.normal_(self.model.fc.weight, 0, 0.01)
            nn.init.zeros_(self.model.fc.bias)

    def train(self, mode: bool = True) -> T:
        super(TorchvisionResNet, self).train(mode=mode)

        if mode and (self.partial_bn or self.fix_bn):
            freezing_bn(self, partial_bn=self.partial_bn)

        return self

    def forward(self, x):
        x = self.model(x)

        return {KEY_OUTPUT: x}


@registry.RECOGNIZER.register('TorchvisionResNet')
def build_torchvision_resnet(cfg):
    torchvision_pretrained = cfg.MODEL.RECOGNIZER.TORCHVISION_PRETRAINED
    pretrained_num_classes = cfg.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES
    fix_bn = cfg.MODEL.NORM.FIX_BN
    partial_bn = cfg.MODEL.NORM.PARTIAL_BN
    # for backbone
    arch = cfg.MODEL.BACKBONE.ARCH
    zero_init_residual = cfg.MODEL.RECOGNIZER.ZERO_INIT_RESIDUAL
    num_classes = cfg.MODEL.HEAD.NUM_CLASSES

    return TorchvisionResNet(
        arch=arch,
        num_classes=num_classes,
        torchvision_pretrained=torchvision_pretrained,
        pretrained_num_classes=pretrained_num_classes,
        fix_bn=fix_bn,
        partial_bn=partial_bn,
        zero_init_residual=zero_init_residual
    )



# -*- coding: utf-8 -*-

"""
@date: 2020/12/4 下午4:11
@file: act_helper.py
@author: zj
@description: 
"""

import numpy as np
import torch
import torch.nn as nn
from .layers.asymmetric_convolution_block import AsymmetricConvolutionBlock
from .layers.acb_util import _fuse_kernel, _add_to_square_kernel
from .layers.repvgg_block import RepVGGBlock
from .layers import repvgg_util
from .layers.diverse_branch_block import DiverseBranchBlock
from .layers import dbb_util


def get_conv(cfg):
    """
    Args:
        cfg (CfgNode): model building configs, details are in the comments of
            the config file.
    Returns:
        nn.Module: the conv layer.
    """
    conv_type = cfg.MODEL.CONV.TYPE
    if conv_type == "Conv2d":
        return nn.Conv2d
    elif conv_type == "Conv3d":
        return nn.Conv3d
    else:
        raise NotImplementedError(
            "Conv type {} is not supported".format(conv_type)
        )


def insert_acblock(model: nn.Module):
    items = list(model.named_children())
    idx = 0
    while idx < len(items):
        name, module = items[idx]
        if isinstance(module, nn.Conv2d) and module.kernel_size[0] > 1:
            # Replace standard convolution with acblock
            in_channels = module.in_channels
            out_channels = module.out_channels
            kernel_size = module.kernel_size
            stride = module.stride
            padding = module.padding
            dilation = module.dilation
            groups = module.groups
            padding_mode = module.padding_mode

            acblock = AsymmetricConvolutionBlock(in_channels,
                                                 out_channels,
                                                 kernel_size[0],
                                                 stride[0],
                                                 padding=padding[0],
                                                 padding_mode=padding_mode,
                                                 dilation=dilation,
                                                 groups=groups)
            model.add_module(name, acblock)
            # If conv layer is followed by BN layer, delete the BN layer
            # refer to [About BN layer #35](https://github.com/DingXiaoH/ACNet/issues/35)
            if (idx + 1) < len(items) and isinstance(items[idx + 1][1], nn.BatchNorm2d):
                new_layer = nn.Identity()
                model.add_module(items[idx + 1][0], new_layer)
        else:
            insert_acblock(module)
        idx += 1


def fuse_acblock(model: nn.Module, eps=1e-5):
    for name, module in model.named_children():
        if isinstance(module, AsymmetricConvolutionBlock):
            # Replace acblock with standard convolution
            # Obtain the weight of NX1 convolution and the weight, offset, runtime mean and runtime variance of corresponding BN
            vertical_conv_weight = module.ver_conv.weight
            vertical_bn_weight = module.ver_bn.weight
            vertical_bn_bias = module.ver_bn.bias
            vertical_bn_running_mean = module.ver_bn.running_mean
            vertical_bn_running_std = torch.sqrt(module.ver_bn.running_var + eps)
            # Obtain the weight of 1xN convolution and the weight, offset, runtime mean and runtime variance of the corresponding BN
            horizontal_conv_weight = module.hor_conv.weight
            horizontal_bn_weight = module.hor_bn.weight
            horizontal_bn_bias = module.hor_bn.bias
            horizontal_bn_running_mean = module.hor_bn.running_mean
            horizontal_bn_running_std = torch.sqrt(module.hor_bn.running_var + eps)
            if isinstance(module.square_bn, nn.Identity):
                square_weight = module.square_conv.weight
                square_bias = module.square_conv.bias
                # Calculation bias
                fused_bias = square_bias + vertical_bn_bias + horizontal_bn_bias \
                             - vertical_bn_running_mean * vertical_bn_weight / vertical_bn_running_std \
                             - horizontal_bn_running_mean * horizontal_bn_weight / horizontal_bn_running_std
                # Calculate weight
                fused_kernel = square_weight
            else:
                square_conv_weight = module.square_conv.weight
                square_bn_weight = module.square_bn.weight
                square_bn_bias = module.square_bn.bias
                square_bn_running_mean = module.square_bn.running_mean
                square_bn_running_std = torch.sqrt(module.square_bn.running_var + eps)
                # Calculate bias
                fused_bias = square_bn_bias + vertical_bn_bias + horizontal_bn_bias \
                             - square_bn_running_mean * square_bn_weight / square_bn_running_std \
                             - vertical_bn_running_mean * vertical_bn_weight / vertical_bn_running_std \
                             - horizontal_bn_running_mean * horizontal_bn_weight / horizontal_bn_running_std
                # Calculate weight
                fused_kernel = _fuse_kernel(square_conv_weight, square_bn_weight, square_bn_running_std)
            # Calculate weight
            _add_to_square_kernel(fused_kernel,
                                  _fuse_kernel(vertical_conv_weight, vertical_bn_weight, vertical_bn_running_std))
            _add_to_square_kernel(fused_kernel,
                                  _fuse_kernel(horizontal_conv_weight, horizontal_bn_weight, horizontal_bn_running_std))
            # Create a new standard convolution, assign weight and bias, and reinsert the model
            fused_conv = nn.Conv2d(module.in_channels,
                                   module.out_channels,
                                   module.kernel_size,
                                   stride=module.stride,
                                   padding=module.padding,
                                   dilation=module.dilation,
                                   groups=module.groups,
                                   padding_mode=module.padding_mode
                                   )
            fused_conv.weight = nn.Parameter(fused_kernel.detach().cpu())
            fused_conv.bias = nn.Parameter(fused_bias.detach().cpu())
            model.add_module(name, fused_conv)
        else:
            fuse_acblock(module, eps=eps)


def insert_repvgg_block(model: nn.Module):
    items = list(model.named_children())
    idx = 0
    while idx < len(items):
        name, module = items[idx]
        if isinstance(module, nn.Conv2d) and module.kernel_size[0] > 1:
            # Replace the standard convolution with repvgg_block
            in_channels = module.in_channels
            out_channels = module.out_channels
            kernel_size = module.kernel_size
            stride = module.stride
            padding = module.padding
            dilation = module.dilation
            groups = module.groups
            padding_mode = module.padding_mode

            acblock = RepVGGBlock(in_channels,
                                  out_channels,
                                  kernel_size[0],
                                  stride[0],
                                  padding=padding[0],
                                  padding_mode=padding_mode,
                                  dilation=dilation,
                                  groups=groups)
            model.add_module(name, acblock)
            # If conv layer is followed by BN layer, delete the BN layer
            # refer to [About BN layer #35](https://github.com/DingXiaoH/ACNet/issues/35)
            if (idx + 1) < len(items) and isinstance(items[idx + 1][1], nn.BatchNorm2d):
                new_layer = nn.Identity()
                model.add_module(items[idx + 1][0], new_layer)
        else:
            insert_repvgg_block(module)
        idx += 1


def fuse_repvgg_block(model: nn.Module):
    for name, module in model.named_children():
        if isinstance(module, RepVGGBlock):
            # Replace repvgg_block with standard convolution
            kernel, bias = repvgg_util.get_equivalent_kernel_bias(module.rbr_dense,
                                                                  module.rbr_1x1,
                                                                  module.rbr_identity,
                                                                  module.in_channels,
                                                                  module.groups,
                                                                  module.padding)
            # Create a new standard convolution, assign weight and bias, and reinsert the model
            fused_conv = nn.Conv2d(module.in_channels,
                                   module.out_channels,
                                   module.kernel_size,
                                   stride=module.stride,
                                   padding=module.padding,
                                   dilation=module.dilation,
                                   groups=module.groups,
                                   padding_mode=module.padding_mode,
                                   bias=True
                                   )
            fused_conv.weight = nn.Parameter(kernel.detach().cpu())
            fused_conv.bias = nn.Parameter(bias.detach().cpu())
            model.add_module(name, fused_conv)
        else:
            fuse_repvgg_block(module)


def insert_dbblock(model: nn.Module):
    items = list(model.named_children())
    idx = 0
    while idx < len(items):
        name, module = items[idx]
        if isinstance(module, nn.Conv2d) and module.kernel_size[0] > 1:
            # Replace the standard convolution with dbblock
            in_channels = module.in_channels
            out_channels = module.out_channels
            kernel_size = module.kernel_size
            stride = module.stride
            padding = module.padding
            dilation = module.dilation
            groups = module.groups

            acblock = DiverseBranchBlock(in_channels,
                                         out_channels,
                                         kernel_size[0],
                                         stride[0],
                                         padding=padding[0],
                                         dilation=dilation[0],
                                         groups=groups)
            model.add_module(name, acblock)
            # If conv layer is followed by BN layer, delete the BN layer
            # refer to [About BN layer #35](https://github.com/DingXiaoH/ACNet/issues/35)
            if (idx + 1) < len(items) and isinstance(items[idx + 1][1], nn.BatchNorm2d):
                new_layer = nn.Identity()
                model.add_module(items[idx + 1][0], new_layer)
        else:
            insert_dbblock(module)
        idx += 1


def fuse_dbblock(model: nn.Module):
    for name, module in model.named_children():
        if isinstance(module, DiverseBranchBlock):
            # Replace dbblock with standard convolution
            kernel, bias = dbb_util.get_equivalent_kernel_bias(module)

            # Create a new standard convolution, assign weight and bias, and reinsert the model
            fused_conv = nn.Conv2d(module.in_channels,
                                   module.out_channels,
                                   module.kernel_size,
                                   stride=module.stride,
                                   padding=module.padding,
                                   dilation=module.dilation,
                                   groups=module.groups,
                                   bias=True
                                   )
            fused_conv.weight = nn.Parameter(kernel.detach().cpu())
            fused_conv.bias = nn.Parameter(bias.detach().cpu())
            model.add_module(name, fused_conv)
        else:
            fuse_dbblock(module)



# -*- coding: utf-8 -*-

"""
@date: 2020/12/28 上午10:35
@file: key_word.py
@author: zj
@description: 
"""

KEY_OUTPUT = 'output'
KEY_LOSS = 'loss'

KEY_IMGS = 'imgs'
KEY_TARGETS = 'targets'
KEY_CLASSES = 'classes'



# -*- coding: utf-8 -*-

"""
@date: 2021/10/3 上午11:17
@file: resize.py
@author: zj
@description: 
"""

import cv2

import albumentations as A
from typing import Sequence


class Resize(object):
    """
    Resize the input to the given size.
    when enlarged image, recommended cv2.INTER_LINEAR; when reduce image, recommended cv2.INTER_ARER.

    Args:
        size (sequence): Desired output size.
        interpolation (OpenCV flag): flag that is used to specify the interpolation algorithm. Should be one of:
            cv2.INTER_NEAREST, cv2.INTER_LINEAR, cv2.INTER_CUBIC, cv2.INTER_AREA, cv2.INTER_LANCZOS4.
            Default: cv2.INTER_LINEAR.
        p (float): probability of applying the transform. Default: 1.

    Image types:
        uint8, float32

    Related:
        torchvision PIL transform always use antialiasing so that's different comparing to other library. See
        https://github.com/pytorch/vision/blob/fbd69f1052292336f78b69804a68ade5a6b2f3b4/torchvision/transforms/functional.py#L357
    """

    def __init__(self, size, interpolation=cv2.INTER_LINEAR, p=1.0):
        if not isinstance(size, Sequence):
            raise TypeError("Size should be int or sequence. Got {}".format(type(size)))
        if isinstance(size, Sequence) and len(size) not in (2, ):
            raise ValueError("If size is a sequence, it should have 2 values")
        self.size = size

        assert interpolation in [cv2.INTER_NEAREST, cv2.INTER_LINEAR, cv2.INTER_CUBIC, cv2.INTER_AREA,
                                 cv2.INTER_LANCZOS4]
        self.interpolation = interpolation
        self.p = p

        new_h, new_w = self.size
        self.t = A.Resize(new_h, new_w, interpolation=self.interpolation, p=self.p)

    def __call__(self, image):
        return self.t(image=image)['image']

    def __repr__(self):
        return self.__class__.__name__ + '(size={0}, interpolation={1}, p={2})'. \
            format(self.size, self.interpolation, self.p)



# -*- coding: utf-8 -*-

"""
@date: 2020/12/24 下午7:38
@file: shufflenetv1.py
@author: zj
@description: 
"""

from zcls.model import registry
from ..base_recognizer import BaseRecognizer


@registry.RECOGNIZER.register('MNASNet')
class MNASNet(BaseRecognizer):

    def __init__(self, cfg):
        super().__init__(cfg)



# -*- coding: utf-8 -*-

"""
@date: 2020/12/4 下午4:11
@file: act_helper.py
@author: zj
@description: 
"""

import torch.nn as nn


def get_act(cfg):
    """
    Args:
        cfg (CfgNode): model building configs, details are in the comments of
            the config file.
    Returns:
        nn.Module: the activation layer.
    """
    act_type = cfg.MODEL.ACT.TYPE
    if act_type == "ReLU":
        return nn.ReLU
    elif act_type == "ReLU6":
        return nn.ReLU6
    elif act_type == 'HSwish':
        return nn.Hardswish
    else:
        raise NotImplementedError(
            "Activation type {} is not supported".format(act_type)
        )


def get_sigmoid(sigmoid_type):
    if sigmoid_type == 'Sigmoid':
        return nn.Sigmoid
    elif sigmoid_type == 'HSigmoid':
        return nn.Hardsigmoid
    else:
        raise IOError(f'{sigmoid_type} doesn\'t exists')



# -*- coding: utf-8 -*-

"""
@date: 2020/8/22 下午9:21
@file: trainer.py
@author: zj
@description: 
"""

import torch

from .cifar import CIFAR
from .fashionmnist import FashionMNIST
from .imagenet import ImageNet
from .general_dataset import GeneralDataset
from .general_dataset_v2 import GeneralDatasetV2
from .lmdb_dataset import LMDBDataset
from .lmdb_imagenet import LMDBImageNet
from .mp_dataset import MPDataset


def build_dataset(cfg, transform=None, target_transform=None, is_train=True, **kwargs):
    dataset_name = cfg.DATASET.NAME
    data_root = cfg.DATASET.TRAIN_ROOT if is_train else cfg.DATASET.TEST_ROOT
    top_k = cfg.DATASET.TOP_K

    if dataset_name == 'CIFAR100':
        dataset = CIFAR(data_root, train=is_train, transform=transform, target_transform=target_transform,
                        top_k=top_k, is_cifar100=True)
    elif dataset_name == 'CIFAR10':
        dataset = CIFAR(data_root, train=is_train, transform=transform, target_transform=target_transform,
                        top_k=top_k, is_cifar100=False)
    elif dataset_name == 'FashionMNIST':
        dataset = FashionMNIST(data_root, train=is_train, transform=transform, target_transform=target_transform,
                               top_k=top_k)
    elif dataset_name == 'ImageNet':
        dataset = ImageNet(data_root, train=is_train, transform=transform, target_transform=target_transform,
                           top_k=top_k)
    elif dataset_name == 'GeneralDataset':
        dataset = GeneralDataset(data_root, transform=transform, target_transform=target_transform, top_k=top_k)
    elif dataset_name == 'GeneralDatasetV2':
        dataset = GeneralDatasetV2(data_root, transform=transform, target_transform=target_transform, top_k=top_k)
    elif dataset_name == 'LMDBDataset':
        dataset = LMDBDataset(data_root, transform=transform, target_transform=target_transform, top_k=top_k)
    elif dataset_name == 'LMDBImageNet':
        dataset = LMDBImageNet(data_root, transform=transform, target_transform=target_transform, top_k=top_k)
    elif dataset_name == 'MPDataset':
        shuffle = cfg.DATALOADER.RANDOM_SAMPLE
        num_gpus = cfg.NUM_GPUS
        rank_id = kwargs.get('rank_id', 0)
        epoch = kwargs.get('epoch', 0)

        dataset = MPDataset(data_root, transform=transform, target_transform=target_transform, top_k=top_k,
                            shuffle=shuffle, num_gpus=num_gpus, rank_id=rank_id, epoch=epoch, drop_last=is_train)
    else:
        raise ValueError(f"the dataset {dataset_name} does not exist")

    return dataset



# -*- coding: utf-8 -*-

"""
@date: 2021/2/2 下午8:51
@file: repvgg_util.py
@author: zj
@description: 
"""

import torch
import torch.nn as nn
import numpy as np


#   This func derives the equivalent kernel and bias in a DIFFERENTIABLE way.
#   You can get the equivalent kernel and bias at any time and do whatever you want,
#   for example, apply some penalties or constraints during training, just like you do to the other models.
#   May be useful for quantization or pruning.
def get_equivalent_kernel_bias(rbr_dense, rbr_1x1, rbr_identity, in_channels, groups, padding_11):
    kernel3x3, bias3x3 = _fuse_bn_tensor(rbr_dense, in_channels, groups)
    kernel1x1, bias1x1 = _fuse_bn_tensor(rbr_1x1, in_channels, groups)
    kernelid, biasid = _fuse_bn_tensor(rbr_identity, in_channels, groups)
    return kernel3x3 + _pad_1x1_to_3x3_tensor(kernel1x1, padding_11) + kernelid, bias3x3 + bias1x1 + biasid


def _pad_1x1_to_3x3_tensor(kernel1x1, padding_11=1):
    if kernel1x1 is None:
        return 0
    else:
        # return torch.nn.functional.pad(kernel1x1, [1, 1, 1, 1])
        return torch.nn.functional.pad(kernel1x1, [padding_11] * 4)


def _fuse_bn_tensor(branch, in_channels, groups):
    if branch is None:
        return 0, 0
    if isinstance(branch, nn.Sequential):
        layer_list = list(branch)
        if len(layer_list) == 2 and isinstance(layer_list[1], nn.Identity):
            # conv/bn已经在acb中进行了融合
            return branch.conv.weight, branch.conv.bias
        kernel = branch.conv.weight
        running_mean = branch.bn.running_mean
        running_var = branch.bn.running_var
        gamma = branch.bn.weight
        beta = branch.bn.bias
        eps = branch.bn.eps
    else:
        assert isinstance(branch, nn.BatchNorm2d)
        input_dim = in_channels // groups
        kernel_value = np.zeros((in_channels, input_dim, 3, 3), dtype=np.float32)
        for i in range(in_channels):
            kernel_value[i, i % input_dim, 1, 1] = 1

        kernel = torch.from_numpy(kernel_value).to(branch.weight.device)
        running_mean = branch.running_mean
        running_var = branch.running_var
        gamma = branch.weight
        beta = branch.bias
        eps = branch.eps
    std = (running_var + eps).sqrt()
    t = (gamma / std).reshape(-1, 1, 1, 1)
    return kernel * t, beta - running_mean * gamma / std



# -*- coding: utf-8 -*-

"""
@date: 2020/11/25 下午6:49
@file: model.py
@author: zj
@description: 
"""

from yacs.config import CfgNode as CN


def add_config(_C):
    # ---------------------------------------------------------------------------- #
    # Model
    # ---------------------------------------------------------------------------- #
    _C.MODEL = CN()

    # ---------------------------------------------------------------------------- #
    # Convolution
    # ---------------------------------------------------------------------------- #
    _C.MODEL.CONV = CN()
    _C.MODEL.CONV.TYPE = 'Conv2d'
    # used in AsymmetricConvolutionBlock/RepVGGBlock/DiverseBranchBlock
    _C.MODEL.CONV.ADD_BLOCKS = None

    # ---------------------------------------------------------------------------- #
    # Normalization
    # ---------------------------------------------------------------------------- #
    _C.MODEL.NORM = CN()
    _C.MODEL.NORM.TYPE = 'BatchNorm2d'
    # batchnorm
    _C.MODEL.NORM.SYNC_BN = False
    _C.MODEL.NORM.FIX_BN = False
    _C.MODEL.NORM.PARTIAL_BN = False
    # Precise BN stats.
    _C.MODEL.NORM.PRECISE_BN = False
    # Number of samples use to compute precise bn.
    _C.MODEL.NORM.NUM_BATCHES_PRECISE = 200
    # groupnorm
    _C.MODEL.NORM.GROUPS = 32

    # ---------------------------------------------------------------------------- #
    # activation
    # ---------------------------------------------------------------------------- #
    _C.MODEL.ACT = CN()
    _C.MODEL.ACT.TYPE = 'ReLU'
    _C.MODEL.ACT.SIGMOID_TYPE = 'Sigmoid'

    # ---------------------------------------------------------------------------- #
    # compression
    # ---------------------------------------------------------------------------- #
    _C.MODEL.COMPRESSION = CN()
    _C.MODEL.COMPRESSION.WIDTH_MULTIPLIER = 1.0
    # Set the number of channels in each layer to be a multiple of 8
    _C.MODEL.COMPRESSION.ROUND_NEAREST = 8

    # ---------------------------------------------------------------------------- #
    # attention
    # ---------------------------------------------------------------------------- #
    _C.MODEL.ATTENTION = CN()
    _C.MODEL.ATTENTION.WITH_ATTENTION = False
    _C.MODEL.ATTENTION.WITH_ATTENTIONS = (0, 0, 0, 0)
    _C.MODEL.ATTENTION.REDUCTION = 16
    _C.MODEL.ATTENTION.ATTENTION_TYPE = 'SqueezeAndExcitationBlock2D'
    # used for se-block
    _C.MODEL.ATTENTION.BIAS = False

    # ---------------------------------------------------------------------------- #
    # backbone
    # ---------------------------------------------------------------------------- #
    _C.MODEL.BACKBONE = CN()
    _C.MODEL.BACKBONE.NAME = 'ShuffleNetV1'
    _C.MODEL.BACKBONE.IN_PLANES = 3
    _C.MODEL.BACKBONE.ARCH = 'resnet18'
    _C.MODEL.BACKBONE.BASE_PLANES = 64
    # base channels each layer
    _C.MODEL.BACKBONE.LAYER_PLANES = (64, 128, 256, 512)
    _C.MODEL.BACKBONE.DOWNSAMPLES = (0, 1, 1, 1)
    # Is AvgPool used instead of Conv2d for downsampling in residual block
    _C.MODEL.BACKBONE.USE_AVG = False
    # Perform downsampling before 3x3
    _C.MODEL.BACKBONE.FAST_AVG = False
    # for MobileNetV1
    _C.MODEL.BACKBONE.STRIDES = (1, 2, 1, 2, 1, 2, 1, 1, 1, 1, 1, 2, 2)
    # for ResNet3D
    _C.MODEL.BACKBONE.CONV1_KERNEL = (1, 7, 7)
    _C.MODEL.BACKBONE.CONV1_STRIDE = (1, 2, 2)
    _C.MODEL.BACKBONE.CONV1_PADDING = (0, 3, 3)
    _C.MODEL.BACKBONE.POOL1_KERNEL = (1, 3, 3)
    _C.MODEL.BACKBONE.POOL1_STRIDE = (1, 2, 2)
    _C.MODEL.BACKBONE.POOL1_PADDING = (0, 1, 1)
    _C.MODEL.BACKBONE.WITH_POOL2 = False
    _C.MODEL.BACKBONE.TEMPORAL_STRIDES = (1, 1, 1, 1)
    _C.MODEL.BACKBONE.INFLATE_LIST = (0, 0, 0, 0)
    _C.MODEL.BACKBONE.INFLATE_STYLE = '3x1x1'
    # for ShuffleNetV1
    _C.MODEL.BACKBONE.WITH_GROUPS = (0, 1, 1)

    # ---------------------------------------------------------------------------- #
    # head
    # ---------------------------------------------------------------------------- #
    _C.MODEL.HEAD = CN()
    _C.MODEL.HEAD.NAME = 'GeneralHead2D'
    _C.MODEL.HEAD.FEATURE_DIMS = 1024
    _C.MODEL.HEAD.DROPOUT_RATE = 0.
    _C.MODEL.HEAD.NUM_CLASSES = 1000
    _C.MODEL.HEAD.BIAS = True
    # for mobilenetv3
    _C.MODEL.HEAD.INNER_DIMS = 1280

    # ---------------------------------------------------------------------------- #
    # recognizer
    # ---------------------------------------------------------------------------- #
    _C.MODEL.RECOGNIZER = CN()
    _C.MODEL.RECOGNIZER.NAME = 'ShuffleNetV1'
    _C.MODEL.RECOGNIZER.TYPE = 'ResNet'
    # zcls training model, used in the test phase
    _C.MODEL.RECOGNIZER.PRELOADED = ""
    # zcls training model, used in the training stage
    _C.MODEL.RECOGNIZER.PRETRAINED = ""
    # torchvision training model, used in the training phase
    _C.MODEL.RECOGNIZER.TORCHVISION_PRETRAINED = False
    # Number of pre-training model categories
    _C.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES = 1000
    # Zero initialization residual connection
    _C.MODEL.RECOGNIZER.ZERO_INIT_RESIDUAL = False

    # ---------------------------------------------------------------------------- #
    # criterion
    # ---------------------------------------------------------------------------- #
    _C.MODEL.CRITERION = CN()
    _C.MODEL.CRITERION.NAME = 'CrossEntropyLoss'
    # for label smoothing loss
    _C.MODEL.CRITERION.SMOOTHING = 0.1
    # mean or sum
    _C.MODEL.CRITERION.REDUCTION = 'mean'



# -*- coding: utf-8 -*-

"""
@date: 2020/8/23 上午9:51
@file: inference.py
@author: zj
@description: 
"""

import os
import datetime
import torch
from tqdm import tqdm

from zcls.config.key_word import KEY_OUTPUT
import zcls.util.logging as logging
from zcls.util.prefetcher import Prefetcher
from zcls.util.distributed import all_gather, is_master_proc

logger = logging.get_logger(__name__)


@torch.no_grad()
def compute_on_dataset(images, targets, model, num_gpus, evaluator):
    output_dict = model(images)
    # Gather all the predictions across all the devices to perform ensemble.
    if num_gpus > 1:
        keys = list()
        values = list()
        keys.append(KEY_OUTPUT)
        values.append(output_dict[KEY_OUTPUT])
        values = all_gather(values)
        output_dict = {k: v for k, v in zip(keys, values)}
        targets = all_gather([targets])[0]

    evaluator.evaluate_test(output_dict, targets)


@torch.no_grad()
def inference(cfg, model, test_data_loader, device, **kwargs):
    cur_epoch = kwargs.get('cur_epoch', None)
    dataset_name = cfg.DATASET.NAME
    num_gpus = cfg.NUM_GPUS

    dataset = test_data_loader.dataset
    evaluator = test_data_loader.dataset.evaluator
    evaluator.clean()

    logger.info("Evaluating {} dataset({} video clips):".format(dataset_name, len(dataset)))

    data_loader = Prefetcher(test_data_loader, device) if cfg.DATALOADER.PREFETCHER else test_data_loader
    if is_master_proc():
        for images, targets in tqdm(data_loader):
            if not cfg.DATALOADER.PREFETCHER:
                images = images.to(device=device, non_blocking=True)
                targets = targets.to(device=device, non_blocking=True)
            compute_on_dataset(images, targets, model, num_gpus, evaluator)
    else:
        for images, targets in data_loader:
            if not cfg.DATALOADER.PREFETCHER:
                images = images.to(device=device, non_blocking=True)
                targets = targets.to(device=device, non_blocking=True)
            compute_on_dataset(images, targets, model, num_gpus, evaluator)

    if cfg.DATALOADER.PREFETCHER:
        data_loader.release()
    result_str, acc_dict = evaluator.get()
    logger.info(result_str)

    if is_master_proc():
        output_dir = cfg.OUTPUT_DIR
        result_path = os.path.join(output_dir,
                                   'result_{}.txt'.format(datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S'))) \
            if cur_epoch is None else os.path.join(output_dir, 'result_{:04d}.txt'.format(cur_epoch))

        with open(result_path, "w") as f:
            f.write(result_str)

    return acc_dict


@torch.no_grad()
def do_evaluation(cfg, model, test_data_loader, device, **kwargs):
    model.eval()

    eval_results = inference(cfg, model, test_data_loader, device, **kwargs)
    torch.cuda.empty_cache()
    return eval_results




# -*- coding: utf-8 -*-

"""
@date: 2020/11/25 下午8:43
@file: batch_norm_2d.py
@author: zj
@description: 
"""

import torch
import torch.nn as nn


class BatchNorm2d(nn.BatchNorm2d):
    """
    from [pytorch_misc/batch_norm_manual.py](https://github.com/ptrblck/pytorch_misc/blob/master/batch_norm_manual.py)
    """

    def __init__(self, num_features, eps=1e-5, momentum=0.1,
                 affine=True, track_running_stats=True):
        super(BatchNorm2d, self).__init__(
            num_features, eps, momentum, affine, track_running_stats)

    def forward(self, input):
        self._check_input_dim(input)

        exponential_average_factor = 0.0

        if self.training and self.track_running_stats:
            if self.num_batches_tracked is not None:
                self.num_batches_tracked += 1
                if self.momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / float(self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self.momentum

        # calculate running estimates
        if self.training:
            mean = input.mean([0, 2, 3])
            # use biased var in train
            var = input.var([0, 2, 3], unbiased=False)
            n = input.numel() / input.size(1)
            with torch.no_grad():
                self.running_mean = exponential_average_factor * mean \
                                    + (1 - exponential_average_factor) * self.running_mean
                # update running_var with unbiased var
                self.running_var = exponential_average_factor * var * n / (n - 1) \
                                   + (1 - exponential_average_factor) * self.running_var
        else:
            mean = self.running_mean
            var = self.running_var

        input = (input - mean[None,:,  None, None]) / (torch.sqrt(var[None, :, None, None] + self.eps))
        if self.affine:
            input = input * self.weight[None, :, None, None] + self.bias[None, :, None, None]

        return input



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""Distributed helpers."""

import functools
import logging
import pickle
import torch
import torch.distributed as dist

LOCAL_PROCESS_GROUP = None
LOCAL_RANK_GROUP = None


def all_gather(tensors):
    """
    All gathers the provided tensors from all processes across machines.
    Args:
        tensors (list): tensors to perform all gather across all processes in
        all machines.
    """

    gather_list = []
    output_tensor = []
    world_size = dist.get_world_size()
    for tensor in tensors:
        tensor_placeholder = [
            torch.ones_like(tensor) for _ in range(world_size)
        ]
        dist.all_gather(tensor_placeholder, tensor, async_op=False)
        gather_list.append(tensor_placeholder)
    for gathered_tensor in gather_list:
        output_tensor.append(torch.cat(gathered_tensor, dim=0))
    return output_tensor


def all_reduce(tensors, average=True):
    """
    All reduce the provided tensors from all processes across machines.
    Args:
        tensors (list): tensors to perform all reduce across all processes in
        all machines.
        average (bool): scales the reduced tensor by the number of overall
        processes across all machines.
    """

    for tensor in tensors:
        dist.all_reduce(tensor, async_op=False)
    if average:
        world_size = dist.get_world_size()
        for tensor in tensors:
            tensor.mul_(1.0 / world_size)
    return tensors


def init_process_group(
        local_rank,
        local_world_size,
        shard_id,
        num_shards,
        init_method,
        dist_backend="nccl",
):
    """
    Initializes the default process group.
    Args:
        local_rank (int): the rank on the current local machine.
        local_world_size (int): the world size (number of processes running) on
        the current local machine.
        shard_id (int): the shard index (machine rank) of the current machine.
        num_shards (int): number of shards for distributed training.
        init_method (string): supporting three different methods for
            initializing process groups:
            "file": use shared file system to initialize the groups across
            different processes.
            "tcp": use tcp address to initialize the groups across different
        dist_backend (string): backend to use for distributed training. Options
            includes gloo, mpi and nccl, the details can be found here:
            https://pytorch.org/docs/stable/distributed.html
    """
    # Sets the GPU to use.
    torch.cuda.set_device(local_rank)
    # Initialize the process group.
    proc_rank = local_rank + shard_id * local_world_size
    world_size = local_world_size * num_shards
    dist.init_process_group(
        backend=dist_backend,
        init_method=init_method,
        world_size=world_size,
        rank=proc_rank,
    )


def is_master_proc(num_gpus=8):
    """
    Determines if the current process is the master process.
    """
    if torch.distributed.is_initialized():
        return dist.get_rank() % num_gpus == 0
    else:
        return True


def is_root_proc():
    """
    Determines if the current process is the root process.
    """
    if torch.distributed.is_initialized():
        return dist.get_rank() == 0
    else:
        return True


def get_world_size():
    """
    Get the size of the world.
    """
    if not dist.is_available():
        return 1
    if not dist.is_initialized():
        return 1
    return dist.get_world_size()


def get_rank():
    """
    Get the rank of the current process.
    """
    if not dist.is_available():
        return 0
    if not dist.is_initialized():
        return 0
    return dist.get_rank()


def synchronize():
    """
    Helper function to synchronize (barrier) among all processes when
    using distributed training
    """
    if not dist.is_available():
        return
    if not dist.is_initialized():
        return
    world_size = dist.get_world_size()
    if world_size == 1:
        return
    dist.barrier()


@functools.lru_cache()
def _get_global_gloo_group():
    """
    Return a process group based on gloo backend, containing all the ranks
    The result is cached.
    Returns:
        (group): pytorch dist group.
    """
    if dist.get_backend() == "nccl":
        return dist.new_group(backend="gloo")
    else:
        return dist.group.WORLD


def _serialize_to_tensor(data, group):
    """
    Seriialize the tensor to ByteTensor. Note that only `gloo` and `nccl`
        backend is supported.
    Args:
        data (data): data to be serialized.
        group (group): pytorch dist group.
    Returns:
        tensor (ByteTensor): tensor that serialized.
    """

    backend = dist.get_backend(group)
    assert backend in ["gloo", "nccl"]
    device = torch.device("cpu" if backend == "gloo" else "cuda")

    buffer = pickle.dumps(data)
    if len(buffer) > 1024 ** 3:
        logger = logging.getLogger(__name__)
        logger.warning(
            "Rank {} trying to all-gather {:.2f} GB of data on device {}".format(
                get_rank(), len(buffer) / (1024 ** 3), device
            )
        )
    storage = torch.ByteStorage.from_buffer(buffer)
    tensor = torch.ByteTensor(storage).to(device=device)
    return tensor


def _pad_to_largest_tensor(tensor, group):
    """
    Padding all the tensors from different GPUs to the largest ones.
    Args:
        tensor (tensor): tensor to pad.
        group (group): pytorch dist group.
    Returns:
        list[int]: size of the tensor, on each rank
        Tensor: padded tensor that has the max size
    """
    world_size = dist.get_world_size(group=group)
    assert (
            world_size >= 1
    ), "comm.gather/all_gather must be called from ranks within the given group!"
    local_size = torch.tensor(
        [tensor.numel()], dtype=torch.int64, device=tensor.device
    )
    size_list = [
        torch.zeros([1], dtype=torch.int64, device=tensor.device)
        for _ in range(world_size)
    ]
    dist.all_gather(size_list, local_size, group=group)
    size_list = [int(size.item()) for size in size_list]

    max_size = max(size_list)

    # we pad the tensor because torch all_gather does not support
    # gathering tensors of different shapes
    if local_size != max_size:
        padding = torch.zeros(
            (max_size - local_size,), dtype=torch.uint8, device=tensor.device
        )
        tensor = torch.cat((tensor, padding), dim=0)
    return size_list, tensor


def all_gather_unaligned(data, group=None):
    """
    Run all_gather on arbitrary picklable data (not necessarily tensors).

    Args:
        data: any picklable object
        group: a torch process group. By default, will use a group which
            contains all ranks on gloo backend.

    Returns:
        list[data]: list of data gathered from each rank
    """
    if get_world_size() == 1:
        return [data]
    if group is None:
        group = _get_global_gloo_group()
    if dist.get_world_size(group) == 1:
        return [data]

    tensor = _serialize_to_tensor(data, group)

    size_list, tensor = _pad_to_largest_tensor(tensor, group)
    max_size = max(size_list)

    # receiving Tensor from all ranks
    tensor_list = [
        torch.empty((max_size,), dtype=torch.uint8, device=tensor.device)
        for _ in size_list
    ]
    dist.all_gather(tensor_list, tensor, group=group)

    data_list = []
    for size, tensor in zip(size_list, tensor_list):
        buffer = tensor.cpu().numpy().tobytes()[:size]
        data_list.append(pickle.loads(buffer))

    return data_list


def init_distributed_training(cfg):
    """
    Initialize variables needed for distributed training.
    """
    if cfg.NUM_GPUS <= 1:
        return
    num_gpus_per_machine = cfg.NUM_GPUS
    num_machines = dist.get_world_size() // num_gpus_per_machine
    for i in range(num_machines):
        ranks_on_i = list(
            range(i * num_gpus_per_machine, (i + 1) * num_gpus_per_machine)
        )
        pg = dist.new_group(ranks_on_i)
        if i == cfg.RANK_ID:
            global LOCAL_PROCESS_GROUP
            LOCAL_PROCESS_GROUP = pg

            global LOCAL_RANK_GROUP
            LOCAL_RANK_GROUP = ranks_on_i


def get_local_size() -> int:
    """
    Returns:
        The size of the per-machine process group,
        i.e. the number of processes per machine.
    """
    if not dist.is_available():
        return 1
    if not dist.is_initialized():
        return 1
    return dist.get_world_size(group=LOCAL_PROCESS_GROUP)


def get_local_rank() -> int:
    """
    Returns:
        The rank of the current process within the local (per-machine) process group.
    """
    if not dist.is_available():
        return 0
    if not dist.is_initialized():
        return 0
    assert LOCAL_PROCESS_GROUP is not None
    return dist.get_rank(group=LOCAL_PROCESS_GROUP)


def get_device(local_rank=None):
    if not torch.cuda.is_available():
        device = torch.device('cpu')
    elif local_rank is None:
        device = torch.cuda.current_device()
    else:
        device = torch.device(f'cuda:{local_rank}')
    return device



import torch
import numpy as np
import torch.nn.functional as F


def transI_fusebn(kernel, bn):
    gamma = bn.weight
    std = (bn.running_var + bn.eps).sqrt()
    return kernel * ((gamma / std).reshape(-1, 1, 1, 1)), bn.bias - bn.running_mean * gamma / std


def transII_addbranch(kernels, biases):
    return sum(kernels), sum(biases)


def transIII_1x1_kxk(k1, b1, k2, b2, groups):
    if groups == 1:
        k = F.conv2d(k2, k1.permute(1, 0, 2, 3))  #
        b_hat = (k2 * b1.reshape(1, -1, 1, 1)).sum((1, 2, 3))
    else:
        k_slices = []
        b_slices = []
        k1_T = k1.permute(1, 0, 2, 3)
        k1_group_width = k1.size(0) // groups
        k2_group_width = k2.size(0) // groups
        for g in range(groups):
            k1_T_slice = k1_T[:, g * k1_group_width:(g + 1) * k1_group_width, :, :]
            k2_slice = k2[g * k2_group_width:(g + 1) * k2_group_width, :, :, :]
            k_slices.append(F.conv2d(k2_slice, k1_T_slice))
            b_slices.append(
                (k2_slice * b1[g * k1_group_width:(g + 1) * k1_group_width].reshape(1, -1, 1, 1)).sum((1, 2, 3)))
        k, b_hat = transIV_depthconcat(k_slices, b_slices)
    return k, b_hat + b2


def transIV_depthconcat(kernels, biases):
    return torch.cat(kernels, dim=0), torch.cat(biases)


def transV_avg(channels, kernel_size, groups):
    input_dim = channels // groups
    k = torch.zeros((channels, input_dim, kernel_size, kernel_size))
    k[np.arange(channels), np.tile(np.arange(input_dim), groups), :, :] = 1.0 / kernel_size ** 2
    return k


#   This has not been tested with non-square kernels (kernel.size(2) != kernel.size(3)) nor even-size kernels
def transVI_multiscale(kernel, target_kernel_size):
    H_pixels_to_pad = (target_kernel_size - kernel.size(2)) // 2
    W_pixels_to_pad = (target_kernel_size - kernel.size(3)) // 2
    return F.pad(kernel, [H_pixels_to_pad, H_pixels_to_pad, W_pixels_to_pad, W_pixels_to_pad])



# -*- coding: utf-8 -*-

"""
@date: 2021/7/28 下午5:51
@file: diver_branch_block.py
@author: zj
@description:
refer to [DingXiaoH/DiverseBranchBlock](https://github.com/DingXiaoH/DiverseBranchBlock)
"""

import torch

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from .. import init_helper


def conv_bn(in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1,
            padding_mode='zeros'):
    conv_layer = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size,
                           stride=(stride, stride), padding=padding, dilation=(dilation, dilation), groups=groups,
                           bias=False, padding_mode=padding_mode)
    bn_layer = nn.BatchNorm2d(num_features=out_channels, affine=True)
    se = nn.Sequential()
    se.add_module('conv', conv_layer)
    se.add_module('bn', bn_layer)
    return se


class IdentityBasedConv1x1(nn.Conv2d):

    def __init__(self, channels, groups=1):
        super(IdentityBasedConv1x1, self).__init__(in_channels=channels, out_channels=channels,
                                                   kernel_size=(1, 1), stride=(1, 1),
                                                   padding=0, groups=groups, bias=False)

        assert channels % groups == 0
        input_dim = channels // groups
        id_value = np.zeros((channels, input_dim, 1, 1))
        for i in range(channels):
            id_value[i, i % input_dim, 0, 0] = 1
        self.id_tensor = torch.from_numpy(id_value).type_as(self.weight)
        nn.init.zeros_(self.weight)

    def forward(self, input):
        kernel = self.weight + self.id_tensor.to(self.weight.device)
        result = F.conv2d(input, kernel, None, stride=1, padding=0, dilation=self.dilation, groups=self.groups)
        return result

    def get_actual_kernel(self):
        return self.weight + self.id_tensor.to(self.weight.device)


class BNAndPadLayer(nn.Module):
    def __init__(self,
                 pad_pixels,
                 num_features,
                 eps=1e-5,
                 momentum=0.1,
                 affine=True,
                 track_running_stats=True):
        super(BNAndPadLayer, self).__init__()
        self.bn = nn.BatchNorm2d(num_features, eps, momentum, affine, track_running_stats)
        self.pad_pixels = pad_pixels

    def forward(self, input):
        output = self.bn(input)
        if self.pad_pixels > 0:
            if self.bn.affine:
                pad_values = self.bn.bias.detach() - self.bn.running_mean * self.bn.weight.detach() / torch.sqrt(
                    self.bn.running_var + self.bn.eps)
            else:
                pad_values = - self.bn.running_mean / torch.sqrt(self.bn.running_var + self.bn.eps)
            output = F.pad(output, [self.pad_pixels] * 4)
            pad_values = pad_values.view(1, -1, 1, 1)
            output[:, :, 0:self.pad_pixels, :] = pad_values
            output[:, :, -self.pad_pixels:, :] = pad_values
            output[:, :, :, 0:self.pad_pixels] = pad_values
            output[:, :, :, -self.pad_pixels:] = pad_values
        return output

    @property
    def weight(self):
        return self.bn.weight

    @property
    def bias(self):
        return self.bn.bias

    @property
    def running_mean(self):
        return self.bn.running_mean

    @property
    def running_var(self):
        return self.bn.running_var

    @property
    def eps(self):
        return self.bn.eps


class DiverseBranchBlock(nn.Module):

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride=1,
                 padding=0,
                 dilation=1,
                 groups=1,
                 internal_channels_1x1_3x3=None,
                 single_init=False):
        super(DiverseBranchBlock, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        assert padding == kernel_size // 2

        self.dbb_origin = conv_bn(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size,
                                  stride=stride, padding=padding, dilation=dilation, groups=groups)

        self.dbb_avg = nn.Sequential()
        if groups < out_channels:
            self.dbb_avg.add_module('conv', nn.Conv2d(in_channels=in_channels, out_channels=out_channels,
                                                      kernel_size=(1, 1), stride=(1, 1),
                                                      padding=0, groups=groups, bias=False))
            self.dbb_avg.add_module('bn', BNAndPadLayer(pad_pixels=padding, num_features=out_channels))
            self.dbb_avg.add_module('avg', nn.AvgPool2d(kernel_size=kernel_size, stride=stride, padding=0))

            # For ordinary convolution, an additional 1x1conv+BN branch is added
            self.dbb_1x1 = conv_bn(in_channels=in_channels, out_channels=out_channels, kernel_size=1, stride=stride,
                                   padding=0, groups=groups)
        else:
            # For depthwise convolution, the pre 1x1conv+BN are cancelled
            self.dbb_avg.add_module('avg', nn.AvgPool2d(kernel_size=kernel_size, stride=stride, padding=padding))

        self.dbb_avg.add_module('avgbn', nn.BatchNorm2d(out_channels))

        if internal_channels_1x1_3x3 is None:
            # For mobilenet, it is better to have 2X internal channels
            internal_channels_1x1_3x3 = in_channels if groups < out_channels else 2 * in_channels

        self.dbb_1x1_kxk = nn.Sequential()
        if internal_channels_1x1_3x3 == in_channels:
            self.dbb_1x1_kxk.add_module('idconv1', IdentityBasedConv1x1(channels=in_channels, groups=groups))
        else:
            self.dbb_1x1_kxk.add_module('conv1', nn.Conv2d(in_channels=in_channels,
                                                           out_channels=internal_channels_1x1_3x3,
                                                           kernel_size=(1, 1), stride=(1, 1), padding=0, groups=groups,
                                                           bias=False))
        self.dbb_1x1_kxk.add_module('bn1', BNAndPadLayer(pad_pixels=padding, num_features=internal_channels_1x1_3x3,
                                                         affine=True))
        self.dbb_1x1_kxk.add_module('conv2',
                                    nn.Conv2d(in_channels=internal_channels_1x1_3x3, out_channels=out_channels,
                                              kernel_size=kernel_size, stride=(stride, stride), padding=0,
                                              groups=groups, bias=False))
        self.dbb_1x1_kxk.add_module('bn2', nn.BatchNorm2d(out_channels))

        self.init_weights()
        #   The experiments reported in the paper used the default initialization of bn.weight (all as 1). But changing the initialization may be useful in some cases.
        if single_init:
            #   Initialize the bn.weight of dbb_origin as 1 and others as 0. This is not the default setting.
            self.single_init()

    def forward(self, inputs):
        out = self.dbb_origin(inputs)
        if hasattr(self, 'dbb_1x1'):
            out += self.dbb_1x1(inputs)
        out += self.dbb_avg(inputs)
        out += self.dbb_1x1_kxk(inputs)
        return out

    def init_weights(self):
        if hasattr(self, "dbb_origin"):
            init_helper.init_weights(self.dbb_origin)
        if hasattr(self, "dbb_1x1"):
            init_helper.init_weights(self.dbb_1x1)
        if hasattr(self, "dbb_avg"):
            init_helper.init_weights(self.dbb_avg)
        if hasattr(self, "dbb_1x1_kxk"):
            init_helper.init_weights(self.dbb_1x1_kxk)

    def init_gamma(self, gamma_value):
        if hasattr(self, "dbb_origin"):
            torch.nn.init.constant_(self.dbb_origin.bn.weight, gamma_value)
        if hasattr(self, "dbb_1x1"):
            torch.nn.init.constant_(self.dbb_1x1.bn.weight, gamma_value)
        if hasattr(self, "dbb_avg"):
            torch.nn.init.constant_(self.dbb_avg.avgbn.weight, gamma_value)
        if hasattr(self, "dbb_1x1_kxk"):
            torch.nn.init.constant_(self.dbb_1x1_kxk.bn2.weight, gamma_value)

    def single_init(self):
        self.init_gamma(0.0)
        if hasattr(self, "dbb_origin"):
            torch.nn.init.constant_(self.dbb_origin.bn.weight, 1.0)



# -*- coding: utf-8 -*-

"""
@date: 2020/11/4 下午1:58
@file: build.py
@author: zj
@description: 
"""

from torch.nn.parallel import DistributedDataParallel as DDP

from zcls.model.conv_helper import insert_acblock, insert_repvgg_block, insert_dbblock
from zcls.model.norm_helper import convert_sync_bn
import zcls.util.distributed as du
from zcls.util.checkpoint import CheckPointer
from zcls.util import logging

from .. import registry
from .shufflenet.shufflenetv1 import ShuffleNetV1
from .shufflenet.shufflenetv2 import ShuffleNetV2
from .shufflenet.torchvision_sfv2 import build_torchvision_sfv2
from .mobilenet.mobilenetv1 import MobileNetV1
from .mobilenet.mobilenetv2 import MobileNetV2
from .mobilenet.torchvision_mobilenetv2 import build_torchvision_mbv2
from .mobilenet.mnasnet import MNASNet
from .mobilenet.torchvision_mnasnet import build_torchvision_mnasnet
from .mobilenet.mobilenetv3 import MobileNetV3
from .mobilenet.torchvision_mobilenetv3 import build_torchvision_mbv3
from .vgg.repvgg import RepVGG
from .resnet.resnet import ResNet
from .resnet.torchvision_resnet import build_torchvision_resnet
from .resnet.official_resnest import build_official_resnest
from .resnet.resnet3d import ResNet3D
from .ghostnet.ghostnet import GhostNet

logger = logging.get_logger(__name__)


def build_recognizer(cfg, device):
    world_size = du.get_world_size()

    model = registry.RECOGNIZER[cfg.MODEL.RECOGNIZER.NAME](cfg)

    if cfg.MODEL.NORM.SYNC_BN and world_size > 1:
        logger.info("start sync BN on the process group of {}".format(du.LOCAL_RANK_GROUP))
        convert_sync_bn(model, du.LOCAL_PROCESS_GROUP)
    if cfg.MODEL.CONV.ADD_BLOCKS is not None:
        assert isinstance(cfg.MODEL.CONV.ADD_BLOCKS, tuple)
        logger.info(f'add blocks: {cfg.MODEL.CONV.ADD_BLOCKS}')
        for add_block in cfg.MODEL.CONV.ADD_BLOCKS:
            if add_block == 'RepVGGBlock':
                insert_repvgg_block(model)
            if add_block == 'ACBlock':
                insert_acblock(model)
            if add_block == 'DiverseBranchBlock':
                insert_dbblock(model)
    preloaded = cfg.MODEL.RECOGNIZER.PRELOADED
    if preloaded != "":
        logger.info(f'load preloaded: {preloaded}')
        check_pointer = CheckPointer(model)
        check_pointer.load(preloaded, map_location=device)
        logger.info("finish loading model weights")

    if du.is_master_proc():
        logger.info(f'full model info:')
        logger.info(model)

    model = model.to(device=device)
    if du.get_world_size() > 1:
        model = DDP(model, device_ids=[device], output_device=device, find_unused_parameters=False)

    return model



# -*- coding: utf-8 -*-

"""
@date: 2020/8/21 下午7:20
@file: build.py
@author: zj
@description: 
"""

from torch.utils.data.distributed import DistributedSampler
from torch.utils.data import RandomSampler

from .datasets.build import build_dataset
from .transforms.build import build_transform
from .dataloader.build import build_dataloader

from .datasets.mp_dataset import MPDataset


def build_data(cfg, is_train=True, **kwargs):
    transform, target_transform = build_transform(cfg, is_train=is_train)
    dataset = build_dataset(cfg, transform=transform, target_transform=target_transform, is_train=is_train, **kwargs)

    return build_dataloader(cfg, dataset, is_train=is_train)


def shuffle_dataset(loader, cur_epoch, is_shuffle=False):
    """"
    Shuffles the data.
    Args:
        loader (loader): data loader to perform shuffle.
        cur_epoch (int): number of the current epoch.
        is_shuffle (bool): need to shuffle the data
    """
    if not is_shuffle:
        return

    dataset = loader.dataset
    if isinstance(dataset, MPDataset):
        dataset.set_epoch(cur_epoch)
    else:
        sampler = loader.sampler
        assert isinstance(
            sampler, (RandomSampler, DistributedSampler)
        ), "Sampler type '{}' not supported".format(type(sampler))
        # RandomSampler handles shuffling automatically
        if isinstance(sampler, DistributedSampler):
            # DistributedSampler shuffles data based on epoch
            sampler.set_epoch(cur_epoch)



# -*- coding: utf-8 -*-

"""
@date: 2020/12/28 上午9:50
@file: place_holder.py
@author: zj
@description: The placeholder layer returns the input data directly without performing any operation
"""
from abc import ABC

import torch.nn as nn


class PlaceHolder(nn.Module, ABC):
    """
    @deprecated. pytorch is implemented, using nn.Identity()
    """

    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x



# -*- coding: utf-8 -*-

"""
@date: 2020/8/27 下午2:28
@file: multistep_lr.py
@author: zj
@description: 
"""

import torch.optim as optim
from torch.optim.optimizer import Optimizer

from .. import registry


@registry.LR_SCHEDULERS.register('MultiStepLR')
def build_multistep_lr(cfg, optimizer):
    assert isinstance(optimizer, Optimizer)

    milestones = cfg.LR_SCHEDULER.MULTISTEP_LR.MILESTONES
    gamma = cfg.LR_SCHEDULER.GAMMA

    return optim.lr_scheduler.MultiStepLR(optimizer, milestones=milestones, gamma=gamma)



# -*- coding: utf-8 -*-

"""
@date: 2020/12/4 上午11:39
@file: mobilenetv2_block.py
@author: zj
@description: 
"""
from abc import ABC

import torch.nn as nn

from .mobilenetv2_inverted_residual import MobileNetV2InvertedResidual


class MobileNetV2Block(nn.Module, ABC):

    def __init__(self,
                 in_channels,
                 out_channels,
                 expansion_rate=1,
                 repeat=1,
                 stride=1,
                 padding=1,
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None,
                 ):
        """
        repeat InvertedResidualUnit. if stride>1, make downsample to the first Unit
        :param in_channels: 输入通道数
        :param out_channels: 输出通道数
        :param expansion_rate: 膨胀因子
        :param repeat: 重复次数
        :param stride: 卷积层步长
        :param padding: 卷积层零填充
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super(MobileNetV2Block, self).__init__()

        features = list()
        for i in range(repeat):
            if i != 0:
                # 仅对第一个残差块执行下采样操作（如果有的话）
                stride = 1
            features.append(MobileNetV2InvertedResidual(in_channels,
                                                        out_channels,
                                                        expansion_rate=expansion_rate,
                                                        stride=stride,
                                                        padding=padding,
                                                        conv_layer=conv_layer,
                                                        norm_layer=norm_layer,
                                                        act_layer=act_layer))
            in_channels = out_channels

        self.conv = nn.Sequential(*features)

    def forward(self, x):
        return self.conv(x)



# -*- coding: utf-8 -*-

"""
@date: 2020/8/21 下午7:55
@file: trainer.py
@author: zj
@description: 
"""

from torch.optim.optimizer import Optimizer

from .. import registry
from .gradual_warmup import GradualWarmupScheduler
from .multistep_lr import build_multistep_lr
from .cosine_annealing_lr import build_cosine_annealing_lr


def build_lr_scheduler(cfg, optimizer):
    assert isinstance(optimizer, Optimizer)
    lr_scheduler = registry.LR_SCHEDULERS[cfg.LR_SCHEDULER.NAME](cfg, optimizer)

    if cfg.LR_SCHEDULER.IS_WARMUP:
        lr_scheduler = GradualWarmupScheduler(optimizer,
                                              multiplier=cfg.LR_SCHEDULER.WARMUP.MULTIPLIER,
                                              total_epoch=cfg.LR_SCHEDULER.WARMUP.ITERATION,
                                              after_scheduler=lr_scheduler)

        optimizer.zero_grad()
        optimizer.step()
        lr_scheduler.step()

    return lr_scheduler



import os
import torch
from torch.nn.parallel import DistributedDataParallel
from torchvision.models.utils import load_state_dict_from_url

from . import logging

logger = logging.get_logger(__name__)


class CheckPointer:
    _last_checkpoint_name = 'last_checkpoint.txt'

    def __init__(self,
                 model,
                 optimizer=None,
                 scheduler=None,
                 save_dir="",
                 save_to_disk=None):
        self.model = model
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.save_dir = save_dir
        self.save_to_disk = save_to_disk

    def save(self, name, **kwargs):
        if not self.save_dir:
            return

        if not self.save_to_disk:
            return

        data = {}
        if isinstance(self.model, DistributedDataParallel):
            data['model'] = self.model.module.state_dict()
        else:
            data['model'] = self.model.state_dict()
        if self.optimizer is not None:
            data["optimizer"] = self.optimizer.state_dict()
        if self.scheduler is not None:
            data["scheduler"] = self.scheduler.state_dict()
        data.update(kwargs)

        save_file = os.path.join(self.save_dir, "{}.pth".format(name))
        logger.info("Saving checkpoint to {}".format(save_file))
        torch.save(data, save_file)

        self.tag_last_checkpoint(save_file)

    def load(self, f=None, use_latest=True, map_location=None):
        if not f and self.has_checkpoint() and use_latest:
            # override argument with existing checkpoint
            f = self.get_checkpoint_file()
        if not f:
            # no checkpoint could be found
            logger.info("No checkpoint found.")
            return {}

        logger.info("Loading checkpoint from {}".format(f))
        checkpoint = self._load_file(f, map_location=map_location)
        model = self.model
        if isinstance(model, DistributedDataParallel):
            model = self.model.module

        model.load_state_dict(checkpoint.pop("model"))
        if "optimizer" in checkpoint and self.optimizer:
            logger.info("Loading optimizer from {}".format(f))
            self.optimizer.load_state_dict(checkpoint.pop("optimizer"))
        if "scheduler" in checkpoint and self.scheduler:
            logger.info("Loading scheduler from {}".format(f))
            self.scheduler.load_state_dict(checkpoint.pop("scheduler"))

        # return any further checkpoint data
        return checkpoint

    def get_checkpoint_file(self):
        save_file = os.path.join(self.save_dir, self._last_checkpoint_name)
        try:
            with open(save_file, "r") as f:
                last_saved = f.read()
                last_saved = last_saved.strip()
        except IOError:
            # if file doesn't exist, maybe because it has just been
            # deleted by a separate process
            last_saved = ""
        return last_saved

    def has_checkpoint(self):
        save_file = os.path.join(self.save_dir, self._last_checkpoint_name)
        return os.path.exists(save_file)

    def tag_last_checkpoint(self, last_filename):
        save_file = os.path.join(self.save_dir, self._last_checkpoint_name)
        with open(save_file, "w") as f:
            f.write(last_filename)

    def _load_file(self, f, map_location=None):
        device = map_location if map_location else torch.device('cpu')
        # local or remote links
        if '://' in f:
            # logger.info(f'load remote url: {f}')
            state_dict = load_state_dict_from_url(f, progress=True, map_location=device)
        else:
            # logger.info(f'load local url: {f}')
            state_dict = torch.load(f, map_location=device)
        return state_dict



# -*- coding: utf-8 -*-

"""
@date: 2021/2/19 下午2:30
@file: misc.py
@author: zj
@description: 
"""

import torch


def make_divisible(v, divisor, min_value=None):
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    :param v:
    :param divisor:
    :param min_value:
    :return:
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


def round_to_multiple_of(val, divisor, round_up_bias=0.9):
    """ Asymmetric rounding to make `val` divisible by `divisor`. With default
    bias, will round up, unless the number is no more than 10% greater than the
    smaller divisible value, i.e. (83, 8) -> 80, but (84, 8) -> 88. """
    assert 0.0 < round_up_bias < 1.0
    new_val = max(divisor, int(val + divisor / 2) // divisor * divisor)
    return new_val if new_val >= round_up_bias * val else new_val + divisor


def channel_shuffle(x, groups):
    """
    # >>> a = torch.arange(12)
    # >>> b = a.reshape(3,4)
    # >>> c = b.transpose(1,0).contiguous()
    # >>> d = c.view(3,4)
    # >>> a
    # tensor([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11])
    # >>> b
    # tensor([[ 0,  1,  2,  3],
    #         [ 4,  5,  6,  7],
    #         [ 8,  9, 10, 11]])
    # >>> c
    # tensor([[ 0,  4,  8],
    #         [ 1,  5,  9],
    #         [ 2,  6, 10],
    #         [ 3,  7, 11]])
    # >>> d
    # tensor([[ 0,  4,  8,  1],
    #         [ 5,  9,  2,  6],
    #         [10,  3,  7, 11]])
    """
    batchsize, num_channels, height, width = x.data.size()
    channels_per_group = num_channels // groups

    # reshape
    x = x.view(batchsize, groups,
               channels_per_group, height, width)

    x = torch.transpose(x, 1, 2).contiguous()

    # flatten
    x = x.view(batchsize, -1, height, width)

    return x



# -*- coding: utf-8 -*-

"""
@date: 2020/12/24 下午8:04
@file: shufflenetv1_unit.py
@author: zj
@description: 
"""

from abc import ABC
import torch
import torch.nn as nn

from ..misc import channel_shuffle


class ShuffleNetV2Unit(nn.Module, ABC):

    def __init__(self,
                 in_channels,
                 out_channels,
                 stride,
                 downsample=None,
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None,
                 ):
        """
        refer to Torchvision realization: https://github.com/pytorch/vision/blob/master/torchvision/models/shufflenetv2.py
        when stride = 1, Unit = Channel Shuffle(Concat(Channel Split(Input), Conv(DWConv(Conv(Channel Split(Input))))));
        when stride = 2, Unit = Channel Shuffle(Concat(Conv(DWConv(Input)), Conv(DWConv(Conv(Input)))));
        official realization (https://github.com/megvii-model/ShuffleNet-Series/blob/master/ShuffleNetV2/blocks.py) has
         a more ingenious implementation, see [A mismatch in shufflenetv1 about ReLU #53](https://github.com/megvii-model/ShuffleNet-Series/issues/53)
        :param in_channels: 输入通道
        :param out_channels: 输出通道
        :param stride: 步长
        :param downsample: 作用于shortcut path
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super().__init__()

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU

        mid_channels = out_channels // 2
        self.conv1 = conv_layer(in_channels, mid_channels, kernel_size=1, stride=1, padding=0, bias=False)
        self.norm1 = norm_layer(mid_channels)

        self.conv2 = conv_layer(mid_channels, mid_channels, kernel_size=3, stride=stride, padding=1, bias=False,
                                groups=mid_channels)
        self.norm2 = norm_layer(mid_channels)

        self.conv3 = conv_layer(mid_channels, mid_channels, kernel_size=1, stride=1, padding=0,
                                bias=False)
        self.norm3 = norm_layer(mid_channels)

        self.act = act_layer(inplace=True)
        self.stride = stride
        self.down_sample = downsample

    def forward(self, x):
        if self.stride == 1:
            x1, x2 = x.chunk(2, dim=1)
        else:
            x1 = x
            x2 = x

        out = self.conv1(x2)
        out = self.norm1(out)
        out = self.act(out)

        out = self.conv2(out)
        out = self.norm2(out)

        out = self.conv3(out)
        out = self.norm3(out)
        out = self.act(out)

        if self.down_sample is not None:
            x1 = self.down_sample(x1)

        out = torch.cat((x1, out), dim=1)

        out = channel_shuffle(out, 2)
        return out



# -*- coding: utf-8 -*-

"""
@date: 2021/10/3 下午12:03
@file: horizontal_flip.py
@author: zj
@description: 
"""

import albumentations as A


class VerticalFlip(object):
    """
    Flip the input vertically around the x-axis.

    Args:
        p (float): probability of applying the transform. Default: 0.5.

    Image types:
        uint8, float32
    """

    def __init__(self, p=0.5):
        self.p = p

        self.t = A.VerticalFlip(p=self.p)

    def __call__(self, image):
        return self.t(image=image)['image']

    def __repr__(self):
        return self.__class__.__name__ + '(p={0})'.format(self.p)



# -*- coding: utf-8 -*-

"""
@date: 2020/12/4 上午11:39
@file: mobilenetv2_inverted_residual.py
@author: zj
@description: MobileNetV2 反向残差块
"""
from abc import ABC

import torch.nn as nn


class MobileNetV2InvertedResidual(nn.Module, ABC):

    def __init__(self,
                 in_channels,
                 out_channels,
                 expansion_rate=1,
                 stride=1,
                 padding=1,
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None,
                 ):
        """
        MobileNetV2的反向残差块由一个膨胀卷积和一个深度可分离卷积(depth-wise conv + point-wise conv)组成
        参考torchvision实现:
        1. 当膨胀率大于1时，执行膨胀卷积操作；
        2. 当深度卷积步长为1且输入/输出通道数相同时，执行残差连接
        3. 反向残差块的最后不执行激活操作
        :param in_channels: 输入通道数
        :param out_channels: 输出通道数
        :param expansion_rate: 膨胀因子
        :param stride: 卷积层步长
        :param padding: 卷积层零填充
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super(MobileNetV2InvertedResidual, self).__init__()

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU6

        # 计算隐藏层输入通道数
        hidden_channels = int(expansion_rate * in_channels)
        features = list()
        if expansion_rate != 1:
            features.append(nn.Sequential(
                conv_layer(in_channels, hidden_channels, kernel_size=1, stride=1, bias=False),
                norm_layer(hidden_channels),
                act_layer(inplace=True)
            ))

        # 深度卷积
        features.append(nn.Sequential(
            conv_layer(hidden_channels, hidden_channels, kernel_size=3, stride=stride, padding=padding, bias=False,
                       groups=hidden_channels),
            norm_layer(hidden_channels),
            act_layer(inplace=True)
        ))

        # 逐点卷积
        features.append(nn.Sequential(
            conv_layer(hidden_channels, out_channels, kernel_size=1, stride=1, bias=False),
            norm_layer(out_channels)
        ))

        self.conv = nn.Sequential(*features)
        self.use_res_connect = stride == 1 and in_channels == out_channels

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)



# -*- coding: utf-8 -*-

"""
@date: 2021/10/4 上午9:40
@file: augment.py
@author: zj
@description: 
"""
import numpy as np
from PIL import Image

import torchvision.transforms.autoaugment as autoaugment
from torchvision.transforms.autoaugment import AutoAugmentPolicy


class AutoAugment(object):
    """
    AutoAugment data augmentation method based on
    `"AutoAugment: Learning Augmentation Strategies from Data" <https://arxiv.org/pdf/1805.09501.pdf>`_.

    Args:
        policy (AutoAugmentPolicy): Desired policy enum defined by
            :class:`torchvision.transforms.autoaugment.AutoAugmentPolicy`. Default is ``AutoAugmentPolicy.IMAGENET``.
        p (float): probability of applying the transform. Default: 0.5.
    """

    def __init__(self, policy: AutoAugmentPolicy = AutoAugmentPolicy.IMAGENET, p=0.5):
        self.policy = policy
        self.p = p

        self.t = autoaugment.AutoAugment(policy=self.policy)

    def __call__(self, image):
        if isinstance(image, np.ndarray):
            image = Image.fromarray(image)

        if np.random.rand(1) < self.p:
            image = self.t(image)

        return np.array(image)

    def __repr__(self):
        return self.__class__.__name__ + '(policy={0}, p={1})'.format(self.policy, self.p)



# -*- coding: utf-8 -*-

"""
@date: 2021/3/18 下午2:13
@file: prefetcher.py
@author: zj
@description: 
"""

import torch
from torch.utils.data import DataLoader

from . import logging

logger = logging.get_logger(__name__)


class Prefetcher():
    """
    from [data_prefetcher](https://github.com/NVIDIA/apex/blob/f5cd5ae937f168c763985f627bbf850648ea5f3f/examples/imagenet/main_amp.py#L256)
    refert to:
    1. [Dose data_prefetcher() really speed up training? #304](https://github.com/NVIDIA/apex/issues/304)
    2. [如何给你PyTorch里的Dataloader打鸡血](https://zhuanlan.zhihu.com/p/66145913)
    Note:
    For overlapped prefetching, supplying pin_memory=True to the dataloader is always required
    """

    def __init__(self, loader: DataLoader, device):
        assert isinstance(loader, DataLoader)
        self.device = device
        self.length = len(loader)
        self.loader = iter(loader)
        self.stream = torch.cuda.Stream()

        self.preload()

    def preload(self):
        try:
            self.next_input, self.next_target = next(self.loader)
        except StopIteration as e:
            self.next_input = None
            self.next_target = None
            return
        with torch.cuda.stream(self.stream):
            # self.next_input = self.next_input.cuda(non_blocking=True)
            # self.next_target = self.next_target.cuda(non_blocking=True)
            self.next_input = self.next_input.to(self.device, non_blocking=True)
            self.next_target = self.next_target.to(self.device, non_blocking=True)

    def __next__(self):
        torch.cuda.current_stream().wait_stream(self.stream)
        if self.next_input is None:
            raise StopIteration
        input = self.next_input
        target = self.next_target
        self.preload()
        return input, target

    def __iter__(self):
        return self

    def __len__(self):
        return self.length

    def release(self):
        del self.stream
        del self.loader
        self.device = None
        self.length = None



# -*- coding: utf-8 -*-

"""
@date: 2021/1/1 下午8:16
@file: selective_kernel_conv2d.py
@author: zj
@description:
"""
from abc import ABC

import torch
import torch.nn as nn
from torch import Tensor

from ..init_helper import init_weights


class SelectiveKernelConv2d(nn.Module, ABC):
    """
    refer to [Selective Kernel Networks](https://arxiv.org/abs/1903.06586)
    """

    def __init__(self,
                 # 输入通道数
                 in_channels: int,
                 # 输出通道数
                 out_channels: int,
                 # 步长
                 stride: int,
                 # 分组数
                 groups: int,
                 # 中间层衰减率
                 reduction_rate: int,
                 # 默认中间层最小通道数
                 default_channels: int = 32,
                 # 维度
                 dimension: int = 2
                 ):
        super().__init__()

        # 3x3 conv
        self.split1 = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, bias=False,
                      padding=1, dilation=1, groups=groups),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )
        # 5x5 conv
        self.split2 = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, bias=False,
                      padding=2, dilation=2, groups=groups),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )
        # fuse
        self.pool = nn.AdaptiveAvgPool2d((1, 1))
        inner_channels = max(out_channels // reduction_rate, default_channels)
        self.compact = nn.Sequential(
            nn.Conv2d(out_channels, inner_channels, kernel_size=1, stride=1, padding=0, bias=False),
            nn.BatchNorm2d(inner_channels),
            nn.ReLU(inplace=True)
        )
        # select
        self.select_a = nn.Linear(inner_channels, out_channels, bias=True)
        self.select_b = nn.Linear(inner_channels, out_channels, bias=True)
        self.softmax = nn.Softmax(dim=0)
        self.dimension = dimension
        self.out_channels = out_channels

        init_weights(self.modules())

    def forward(self, x: Tensor) -> Tensor:
        N, C = x.shape[:2]

        # split
        u1 = self.split1(x)
        u2 = self.split2(x)
        # fuse
        u = u1 + u2
        s = self.pool(u)
        z = self.compact(s).flatten(1)
        # select
        # ac = torch.matmul(z, self.select_a)
        # bc = torch.matmul(z, self.select_b)
        ac = self.select_a(z)
        bc = self.select_b(z)
        abc = torch.stack((ac, bc)).reshape(2, N, self.out_channels, *([1] * self.dimension))
        softmax_abc = self.softmax(abc)
        v = torch.sum(torch.stack((u1, u2)).mul(softmax_abc), dim=0)

        return v



# -*- coding: utf-8 -*-

"""
@date: 2020/12/30 下午4:44
@file: attention_helper.py
@author: zj
@description: 
"""

from zcls.model.layers.global_context_block import GlobalContextBlock2D
from zcls.model.layers.squeeze_and_excitation_block import SqueezeAndExcitationBlock2D
from zcls.model.layers.non_local_embedded_gaussian import NonLocal2DEmbeddedGaussian
from zcls.model.layers.simplified_non_local_embedded_gaussian import SimplifiedNonLocal2DEmbeddedGaussian


def make_attention_block(in_planes, reduction, attention_type, **kwargs):
    if attention_type == 'GlobalContextBlock2D':
        return GlobalContextBlock2D(in_channels=in_planes, reduction=reduction)
    elif attention_type == 'SqueezeAndExcitationBlock2D':
        return SqueezeAndExcitationBlock2D(in_channels=in_planes, reduction=reduction, **kwargs)
    elif attention_type == 'NonLocal2DEmbeddedGaussian':
        return NonLocal2DEmbeddedGaussian(in_channels=in_planes)
    elif attention_type == 'SimplifiedNonLocal2DEmbeddedGaussian':
        return SimplifiedNonLocal2DEmbeddedGaussian(in_channels=in_planes)
    else:
        raise ValueError('no matching type')



# -*- coding: utf-8 -*-

"""
@date: 2021/4/4 下午2:55
@file: general_dataset.py
@author: zj
@description: 
"""

import six
import os
import lmdb
import pickle
from PIL import Image
from torch.utils.data import Dataset

# [Python Pillow - ValueError: Decompressed Data Too Large](https://stackoverflow.com/questions/42671252/python-pillow-valueerror-decompressed-data-too-large)
from PIL import PngImagePlugin

LARGE_ENOUGH_NUMBER = 100
PngImagePlugin.MAX_TEXT_CHUNK = LARGE_ENOUGH_NUMBER * (1024 ** 2)

# [Logs polluted by PIL.PngImagePlugin DEBUG log-level #15](https://github.com/camptocamp/pytest-odoo/issues/15)
import logging

pil_logger = logging.getLogger('PIL')
pil_logger.setLevel(logging.INFO)

from .evaluator.general_evaluator import GeneralEvaluator


def load_data(buf):
    """
    Args:
        buf: the output of `dumps`.
    """
    return pickle.loads(buf)


class LMDBDataset(Dataset):
    """
    [What is the meta.bin file used by the ImageNet dataset? #1646](https://github.com/pytorch/vision/issues/1646)
    torchvision will parse the devkit archive of the ImageNet2012 classification dataset and save
    the meta information in a binary file.
    about problem: TypeError: can't pickle Environment objects
    refert to [TypeError: can't pickle Environment objects when num_workers > 0 for LSUN #689](https://github.com/pytorch/vision/issues/689)
    """

    def __init__(self, root, transform=None, target_transform=None, top_k=(1, 5)):
        assert os.path.isfile(root)

        self.dbpath = root
        self.transform = transform
        self.target_transform = target_transform
        # create evaluator
        self._update_evaluator(top_k)

    def __getitem__(self, index: int):
        if not hasattr(self, 'txn'):
            self.open_lmdb()
        # with env.begin(write=False) as txn:
        byteflow = self.txn.get(self.keys[index])

        imgbuf, target = load_data(byteflow)
        image = self.get_image(imgbuf)

        if self.transform is not None:
            image = self.transform(image)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return image, target

    def __len__(self) -> int:
        if not hasattr(self, 'txn'):
            self.open_lmdb()
        return self.length

    def _update_evaluator(self, top_k):
        if not hasattr(self, 'txn'):
            self.open_lmdb()
        self.evaluator = GeneralEvaluator(self.classes, top_k=top_k)

    def __repr__(self):
        return self.__class__.__name__ + ' (' + self.dbpath + ')'

    def open_lmdb(self):
        self.env = lmdb.open(self.dbpath, subdir=False,
                             readonly=True, lock=False,
                             readahead=False, meminit=False)
        self.txn = self.env.begin(write=False, buffers=True)
        self.length = load_data(self.txn.get(b'__len__'))
        self.keys = load_data(self.txn.get(b'__keys__'))
        self.classes = load_data(self.txn.get(b'classes'))

    def get_classes(self):
        if not hasattr(self, 'txn'):
            self.open_lmdb()
        return self.classes

    def get_image(self, imgbuf):
        buf = six.BytesIO()
        buf.write(imgbuf)
        buf.seek(0)

        image = Image.open(buf).convert('RGB')
        return image



# -*- coding: utf-8 -*-

"""
@date: 2020/12/8 下午3:08
@file: non_local_embedded_gaussian.py
@author: zj
@description: 
"""

import torch
import torch.nn as nn


class _SimplifiedNonLocalNDEmbeddedGaussian(nn.Module):

    def __init__(self,
                 in_channels,
                 dimension=2,
                 norm_layer=None,
                 zero_init_final_norm=True
                 ):
        """
        refer to [context_block.py](https://github.com/xvjiarui/GCNet/blob/master/mmdet/ops/gcb/context_block.py)
        :param in_channels: Input channels
        :param dimension: Data dimension
        :param norm_layer: Normalized layer type
        :param zero_init_final_norm: Initialize the last BN layer at zero to ensure the consistent connection of the initial model
        """
        super(_SimplifiedNonLocalNDEmbeddedGaussian, self).__init__()
        assert dimension in [1, 2, 3]

        self.in_channels = in_channels
        self.dimension = dimension
        self.norm_layer = norm_layer
        self.zero_init_final_norm = zero_init_final_norm

        if self.dimension == 1:
            self.conv_layer = nn.Conv1d
            if self.norm_layer is None:
                self.norm_layer = nn.BatchNorm1d
        elif self.dimension == 2:
            self.conv_layer = nn.Conv2d
            if self.norm_layer is None:
                self.norm_layer = nn.BatchNorm2d
        else:
            self.conv_layer = nn.Conv3d
            if self.norm_layer is None:
                self.norm_layer = nn.BatchNorm3d

        self._construct_nonlocal()
        self.init_weights()

    def _construct_nonlocal(self):
        self.w_k = self.conv_layer(self.in_channels, 1, kernel_size=1, stride=1, padding=0)
        self.w_v = nn.Sequential(
            self.conv_layer(self.in_channels, self.in_channels, kernel_size=1, stride=1, padding=0),
            self.norm_layer(self.in_channels)
        )

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.GroupNorm)):
                if self.zero_init_final_norm:
                    nn.init.constant_(m.weight, 0)
                else:
                    nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        """
        :param x: (N, C, **)
        """
        self._check_input_dim(x)

        identity = x
        N, C = x.shape[:2]

        # (N, C, **) -> (N, C, D_j) -> (N, 1, C, D_j)
        input_x = x.view(N, C, -1).unsqueeze(1)
        # (N, C, **) -> (N, 1, **) -> (N, 1, D_j) -> (N, 1, D_j, 1)
        context_mask = self.w_k(x).view(N, 1, -1).unsqueeze(-1)
        context_mask = nn.functional.softmax(context_mask, dim=2)
        # (N, 1, C, D_j) * (N, 1, D_j, 1) -> (N, 1, C, 1) -> (N, C) -> (N, C, **)
        context = torch.matmul(input_x, context_mask).reshape(N, C).reshape(N, C, *([1] * self.dimension))

        transform = self.w_v(context)

        z = transform + identity
        return z

    def _check_input_dim(self, input):
        raise NotImplementedError


class SimplifiedNonLocal1DEmbeddedGaussian(_SimplifiedNonLocalNDEmbeddedGaussian):

    def __init__(self, in_channels, dimension=1, norm_layer=None, zero_init_final_norm=True):
        super().__init__(in_channels, dimension, norm_layer, zero_init_final_norm)

    def _check_input_dim(self, input):
        if input.dim() != 2 and input.dim() != 3:
            raise ValueError('expected 2D or 3D input (got {}D input)'
                             .format(input.dim()))


class SimplifiedNonLocal2DEmbeddedGaussian(_SimplifiedNonLocalNDEmbeddedGaussian):

    def __init__(self, in_channels, dimension=2, norm_layer=None, zero_init_final_norm=True):
        super().__init__(in_channels, dimension, norm_layer, zero_init_final_norm)

    def _check_input_dim(self, input):
        if input.dim() != 4:
            raise ValueError('expected 4D input (got {}D input)'
                             .format(input.dim()))


class SimplifiedNonLocal3DEmbeddedGaussian(_SimplifiedNonLocalNDEmbeddedGaussian):

    def __init__(self, in_channels, dimension=3, norm_layer=None, zero_init_final_norm=True):
        super().__init__(in_channels, dimension, norm_layer, zero_init_final_norm)

    def _check_input_dim(self, input):
        if input.dim() != 5:
            raise ValueError('expected 5D input (got {}D input)'
                             .format(input.dim()))
